"""Cloud exchange configuration file parsing."""

from __future__ import annotations

import logging
import pathlib
import sys
from typing import Any
from typing import BinaryIO
from typing import Dict  # noqa: UP035
from typing import Literal
from typing import Optional
from typing import TypeVar
from typing import Union

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import Self
else:  # pragma: <3.11 cover
    from typing_extensions import Self

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    import tomllib
else:  # pragma: <3.11 cover
    import tomli as tomllib  # type: ignore[import-not-found,unused-ignore]


BaseModelT = TypeVar('BaseModelT', bound=BaseModel)


class ExchangeAuthConfig(BaseModel):
    """Exchange authentication configuration.

    Attributes:
        method: Authentication method.
        kwargs: Arbitrary keyword arguments to pass to the authenticator.
            The kwargs are excluded from the [`repr()`][repr] of this
            class because they often contain secrets.
    """

    model_config = ConfigDict(extra='forbid')

    method: Optional[Literal['globus']] = None  # noqa: UP045
    kwargs: Dict[str, Any] = Field(  # noqa: UP006
        default_factory=dict,
        repr=False,
    )


class ExchangeServingConfig(BaseModel):
    """Exchange serving configuration.

    Attributes:
        host: Network interface the server binds to.
        port: Network port the server binds to.
        certfile: Certificate file (PEM format) use to enable TLS.
        keyfile: Private key file. If not specified, the key will be
            taken from the certfile.
        auth: Authentication configuration.
        log_file: Location to write logs.
        log_level: Verbosity of logs.
    """

    host: str = 'localhost'
    port: int = 8700
    certfile: Optional[str] = None  # noqa: UP045
    keyfile: Optional[str] = None  # noqa: UP045
    auth: ExchangeAuthConfig = Field(default_factory=ExchangeAuthConfig)
    log_file: Optional[str] = None  # noqa: UP045
    log_level: Union[int, str] = logging.INFO  # noqa: UP007

    @classmethod
    def from_toml(cls, filepath: str | pathlib.Path) -> Self:
        """Parse an TOML config file.

        Example:
            Minimal config without SSL and without authentication.
            ```toml title="exchange.toml"
            port = 8700
            ```

            ```python
            from academy.exchange.cloud.config import ExchangeServingConfig

            config = ExchangeServingConfig.from_toml('exchange.toml')
            ```

        Example:
            Serve with SSL and Globus Auth.
            ```toml title="relay.toml"
            host = "0.0.0.0"
            port = 8700
            certfile = "/path/to/cert.pem"
            keyfile = "/path/to/privkey.pem"

            [auth]
            method = "globus"

            [auth.kwargs]
            client_id = "..."
            client_secret = "..."
            ```

        Note:
            Omitted values will be set to their defaults (if they are an
            optional value with a default).
            ```toml title="relay.toml"
            [serving]
            certfile = "/path/to/cert.pem"
            ```

            ```python
            from academy.exchange.cloud.config import ExchangeServingConfig

            config = ExchangeServingConfig.from_config('relay.toml')
            assert config.certfile == '/path/to/cert.pem'
            assert config.keyfile is None
            ```
        """
        with open(filepath, 'rb') as f:
            return load(cls, f)


def load(model: type[BaseModelT], fp: BinaryIO) -> BaseModelT:
    """Parse TOML from a binary file to a data class.

    Args:
        model: Config model type to parse TOML using.
        fp: File-like bytes stream to read in.

    Returns:
        Model initialized from TOML file.
    """
    return loads(model, fp.read().decode())


def loads(model: type[BaseModelT], data: str) -> BaseModelT:
    """Parse TOML string to data class.

    Args:
        model: Config model type to parse TOML using.
        data: TOML string to parse.

    Returns:
        Model initialized from TOML file.
    """
    data_dict = tomllib.loads(data)
    return model.model_validate(data_dict, strict=True)



from __future__ import annotations

import asyncio
import logging
import time
from concurrent.futures import ThreadPoolExecutor
from typing import Any

import pytest

from academy.agent import action
from academy.agent import Agent
from academy.agent import event
from academy.agent import loop
from academy.agent import timer
from academy.context import ActionContext
from academy.context import AgentContext
from academy.exception import AgentNotInitializedError
from academy.exchange import LocalExchangeTransport
from academy.exchange import UserExchangeClient
from academy.identifier import AgentId
from testing.agents import EmptyAgent
from testing.agents import IdentityAgent
from testing.agents import WaitAgent
from testing.constant import TEST_SLEEP_INTERVAL
from testing.constant import TEST_WAIT_TIMEOUT


def test_initialize_base_type_error() -> None:
    error = 'The Agent type cannot be instantiated directly'
    with pytest.raises(TypeError, match=error):
        Agent()


@pytest.mark.asyncio
async def test_agent_context_initialized_ok(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    agent = EmptyAgent()

    async def _handler(_: Any) -> None:  # pragma: no cover
        pass

    registration = await exchange_client.register_agent(EmptyAgent)
    factory = exchange_client.factory()
    async with await factory.create_agent_client(
        registration,
        _handler,
    ) as client:
        context = AgentContext(
            agent_id=client.client_id,
            exchange_client=client,
            executor=None,  # type: ignore[arg-type]
            shutdown_event=asyncio.Event(),
        )
        agent._agent_set_context(context)

        assert agent.agent_context is context
        assert agent.agent_id is context.agent_id
        assert agent.agent_exchange_client is context.exchange_client

        agent.agent_shutdown()
        assert context.shutdown_event.is_set()


@pytest.mark.asyncio
async def test_agent_context_initialized_error() -> None:
    agent = EmptyAgent()

    with pytest.raises(AgentNotInitializedError):
        _ = agent.agent_context
    with pytest.raises(AgentNotInitializedError):
        _ = agent.agent_id
    with pytest.raises(AgentNotInitializedError):
        _ = agent.agent_exchange_client
    with pytest.raises(AgentNotInitializedError):
        agent.agent_shutdown()


@pytest.mark.asyncio
async def test_agent_run_sync() -> None:
    class SyncAgent(Agent):
        def add_sync(self, a: int, b: int) -> int:
            return a + b

        async def add_async(self, a: int, b: int) -> int:
            return await self.agent_run_sync(self.add_sync, a, b)

    agent = SyncAgent()
    with ThreadPoolExecutor() as executor:
        context: AgentContext[SyncAgent] = AgentContext(
            agent_id=AgentId.new(),
            exchange_client=None,  # type: ignore[arg-type]
            executor=executor,
            shutdown_event=asyncio.Event(),
        )
        agent._agent_set_context(context)

        assert await agent.add_async(0, 1) == 1


@pytest.mark.asyncio
async def test_agent_run_sync_overloaded_warning(caplog) -> None:
    agent = EmptyAgent()
    with ThreadPoolExecutor(max_workers=1) as executor:
        context: AgentContext[EmptyAgent] = AgentContext(
            agent_id=AgentId.new(),
            exchange_client=None,  # type: ignore[arg-type]
            executor=executor,
            shutdown_event=asyncio.Event(),
        )
        agent._agent_set_context(context)

        with caplog.at_level(logging.WARNING):
            tasks = tuple(
                asyncio.create_task(
                    agent.agent_run_sync(time.sleep, TEST_SLEEP_INTERVAL),
                )
                for _ in range(8)
            )
            await asyncio.wait(tasks)

        assert 'sync function "sleep" is waiting for a worker' in caplog.text


@pytest.mark.asyncio
async def test_agent_empty() -> None:
    agent = EmptyAgent()
    await agent.agent_on_startup()

    assert isinstance(agent, EmptyAgent)
    assert isinstance(str(agent), str)
    assert isinstance(repr(agent), str)

    assert len(agent._agent_actions()) == 0
    assert len(agent._agent_loops()) == 0

    await agent.agent_on_shutdown()


@pytest.mark.asyncio
async def test_agent_ignore_property_attributes() -> None:
    class Example(Agent):
        @property
        def bad(self) -> str:  # pragma: no cover
            raise RuntimeError('Property was accessed!')

    agent = Example()
    attributes = set(agent._agent_attributes())
    assert len(attributes) == 0


@pytest.mark.asyncio
async def test_agent_actions() -> None:
    agent = IdentityAgent()
    await agent.agent_on_startup()

    actions = agent._agent_actions()
    assert set(actions) == {'identity'}

    assert await agent.identity(1) == 1

    await agent.agent_on_shutdown()


@pytest.mark.asyncio
async def test_agent_loops() -> None:
    agent = WaitAgent()
    await agent.agent_on_startup()

    loops = agent._agent_loops()
    assert set(loops) == {'wait'}

    shutdown = asyncio.Event()
    shutdown.set()
    await agent.wait(shutdown)

    await agent.agent_on_shutdown()


@pytest.mark.asyncio
async def test_agent_event() -> None:
    class _Event(Agent):
        def __init__(self) -> None:
            self.event = asyncio.Event()
            self.ran = asyncio.Event()
            self.bad = 42

        @event('event')
        async def run(self) -> None:
            self.ran.set()

        @event('missing')
        async def missing_event(self) -> None: ...

        @event('bad')
        async def bad_event(self) -> None: ...

    agent = _Event()

    loops = agent._agent_loops()
    assert set(loops) == {'bad_event', 'missing_event', 'run'}

    shutdown = asyncio.Event()

    with pytest.raises(AttributeError, match='missing'):
        await agent.missing_event(shutdown)
    with pytest.raises(TypeError, match='bad'):
        await agent.bad_event(shutdown)

    task: asyncio.Task[None] = asyncio.create_task(agent.run(shutdown))

    for _ in range(5):
        assert not agent.ran.is_set()
        agent.event.set()
        await asyncio.wait_for(agent.ran.wait(), timeout=TEST_WAIT_TIMEOUT)
        agent.ran.clear()

    shutdown.set()
    await asyncio.wait_for(task, timeout=TEST_WAIT_TIMEOUT)


@pytest.mark.asyncio
async def test_agent_timer() -> None:
    class _Timer(Agent):
        def __init__(self) -> None:
            self.count = 0

        @timer(TEST_SLEEP_INTERVAL)
        async def counter(self) -> None:
            self.count += 1

    agent = _Timer()

    loops = agent._agent_loops()
    assert set(loops) == {'counter'}

    shutdown = asyncio.Event()
    task: asyncio.Task[None] = asyncio.create_task(agent.counter(shutdown))

    await asyncio.sleep(TEST_SLEEP_INTERVAL * 10)
    shutdown.set()

    await asyncio.wait_for(task, timeout=TEST_WAIT_TIMEOUT)


def test_agent_action_decorator_usage_ok() -> None:
    class _TestAgent(Agent):
        @action
        async def action1(self) -> None: ...

        @action()
        async def action2(self) -> None: ...

        @action(context=True)
        async def action3(self, *, context: ActionContext) -> None: ...

    agent = _TestAgent()
    assert len(agent._agent_actions()) == 3  # noqa: PLR2004


def test_agent_action_decorator_usage_error() -> None:
    class _TestAgent(Agent):
        async def missing_arg(self) -> None: ...
        async def pos_only(self, context: ActionContext, /) -> None: ...

    with pytest.raises(
        TypeError,
        match='Action method "missing_arg" must accept a "context"',
    ):
        action(context=True)(_TestAgent.missing_arg)

    with pytest.raises(
        TypeError,
        match='The "context" argument to action method "pos_only"',
    ):
        action(context=True)(_TestAgent.pos_only)


def test_agent_action_decorator_name_clash_ok() -> None:
    class _TestAgent(Agent):
        async def ping(self) -> None: ...

    action(allow_protected_name=True)(_TestAgent.ping)


def test_agent_action_decorator_name_clash_error() -> None:
    class _TestAgent(Agent):
        async def action(self) -> None: ...
        async def ping(self) -> None: ...
        async def shutdown(self) -> None: ...

    with pytest.warns(
        UserWarning,
        match='The name of the decorated method is "action" which clashes',
    ):
        action(_TestAgent.action)

    with pytest.warns(
        UserWarning,
        match='The name of the decorated method is "ping" which clashes',
    ):
        action(_TestAgent.ping)

    with pytest.warns(
        UserWarning,
        match='The name of the decorated method is "shutdown" which clashes',
    ):
        action(_TestAgent.shutdown)


class A(Agent): ...


class B(Agent): ...


class C(A): ...


class D(A, B): ...


def test_agent_mro() -> None:
    assert Agent._agent_mro() == ()
    assert A._agent_mro() == (f'{__name__}.A',)
    assert B._agent_mro() == (f'{__name__}.B',)
    assert C._agent_mro() == (f'{__name__}.C', f'{__name__}.A')
    assert D._agent_mro() == (
        f'{__name__}.D',
        f'{__name__}.A',
        f'{__name__}.B',
    )


def test_invalid_loop_signature() -> None:
    class BadAgent(Agent):
        async def loop(self) -> None: ...

    with pytest.raises(TypeError, match='Signature of loop method "loop"'):
        loop(BadAgent.loop)  # type: ignore[arg-type]



"""HTTP message exchange client and server.

To start the exchange:
```bash
python -m academy.exchange.cloud --config exchange.yaml
```

Connect to the exchange through the client.
```python
from academy.exchange import HttpExchangeFactory

with HttpExchangeFactory(
    'http://localhost:1234'
).create_user_client() as exchange:
    aid, agent_info = exchange.register_agent()
    ...
```
"""

from __future__ import annotations

import argparse
import asyncio
import contextlib
import enum
import logging
import ssl
import sys
import uuid
from collections.abc import Awaitable
from collections.abc import Sequence
from typing import Any
from typing import Callable

if sys.version_info >= (3, 13):  # pragma: >=3.13 cover
    from asyncio import Queue
    from asyncio import QueueEmpty
    from asyncio import QueueShutDown

    AsyncQueue = Queue
else:  # pragma: <3.13 cover
    # Use of queues here is isolated to a single thread/event loop so
    # we only need culsans queues for the backport of shutdown() agent
    from culsans import AsyncQueue
    from culsans import AsyncQueueEmpty as QueueEmpty
    from culsans import AsyncQueueShutDown as QueueShutDown
    from culsans import Queue

from aiohttp.web import AppKey
from aiohttp.web import Application
from aiohttp.web import json_response
from aiohttp.web import middleware
from aiohttp.web import Request
from aiohttp.web import Response
from aiohttp.web import run_app
from pydantic import TypeAdapter
from pydantic import ValidationError

from academy.exception import BadEntityIdError
from academy.exception import MailboxTerminatedError
from academy.exchange.cloud.authenticate import Authenticator
from academy.exchange.cloud.authenticate import get_authenticator
from academy.exchange.cloud.config import ExchangeAuthConfig
from academy.exchange.cloud.config import ExchangeServingConfig
from academy.exchange.cloud.exceptions import ForbiddenError
from academy.exchange.cloud.exceptions import UnauthorizedError
from academy.exchange.transport import MailboxStatus
from academy.identifier import AgentId
from academy.identifier import EntityId
from academy.logging import init_logging
from academy.message import ErrorResponse
from academy.message import Message

logger = logging.getLogger(__name__)


class StatusCode(enum.Enum):
    """Http status codes."""

    OKAY = 200
    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    FORBIDDEN = 403
    NOT_FOUND = 404
    TIMEOUT = 408
    TERMINATED = 419
    NO_RESPONSE = 444


class _MailboxManager:
    def __init__(self) -> None:
        self._owners: dict[EntityId, str | None] = {}
        self._mailboxes: dict[EntityId, AsyncQueue[Message[Any]]] = {}
        self._terminated: set[EntityId] = set()
        self._agents: dict[AgentId[Any], tuple[str, ...]] = {}
        self._locks: dict[EntityId, asyncio.Lock] = {}

    def has_permissions(
        self,
        client: str | None,
        entity: EntityId,
    ) -> bool:
        return entity not in self._owners or self._owners[entity] == client

    async def check_mailbox(
        self,
        client: str | None,
        uid: EntityId,
    ) -> MailboxStatus:
        if uid not in self._mailboxes:
            return MailboxStatus.MISSING
        elif not self.has_permissions(client, uid):
            raise ForbiddenError(
                'Client does not have correct permissions.',
            )

        async with self._locks[uid]:
            if uid in self._terminated:
                return MailboxStatus.TERMINATED
            else:
                return MailboxStatus.ACTIVE

    def create_mailbox(
        self,
        client: str | None,
        uid: EntityId,
        agent: tuple[str, ...] | None = None,
    ) -> None:
        if not self.has_permissions(client, uid):
            raise ForbiddenError(
                'Client does not have correct permissions.',
            )

        mailbox = self._mailboxes.get(uid, None)
        if mailbox is None:
            if sys.version_info >= (3, 13):  # pragma: >=3.13 cover
                queue: AsyncQueue[Message[Any]] = Queue()
            else:  # pragma: <3.13 cover
                queue: AsyncQueue[Message[Any]] = Queue().async_q
            self._mailboxes[uid] = queue
            self._terminated.discard(uid)
            self._owners[uid] = client
            self._locks[uid] = asyncio.Lock()
            if agent is not None and isinstance(uid, AgentId):
                self._agents[uid] = agent
            logger.info('Created mailbox for %s', uid)

    async def terminate(self, client: str | None, uid: EntityId) -> None:
        if not self.has_permissions(client, uid):
            raise ForbiddenError(
                'Client does not have correct permissions.',
            )

        self._terminated.add(uid)
        mailbox = self._mailboxes.get(uid, None)
        if mailbox is None:
            return

        async with self._locks[uid]:
            messages = await _drain_queue(mailbox)
            for message in messages:
                if message.is_request():
                    error = MailboxTerminatedError(uid)
                    body = ErrorResponse(exception=error)
                    response = message.create_response(body)
                    with contextlib.suppress(Exception):
                        await self.put(client, response)

            mailbox.shutdown(immediate=True)
            logger.info('Closed mailbox for %s', uid)

    async def discover(
        self,
        client: str | None,
        agent: str,
        allow_subclasses: bool,
    ) -> list[AgentId[Any]]:
        found: list[AgentId[Any]] = []
        for aid, agents in self._agents.items():
            if not self.has_permissions(client, aid):
                continue
            if aid in self._terminated:
                continue
            if agent == agents[0] or (allow_subclasses and agent in agents):
                found.append(aid)
        return found

    async def get(
        self,
        client: str | None,
        uid: EntityId,
        *,
        timeout: float | None = None,
    ) -> Message[Any]:
        if not self.has_permissions(client, uid):
            raise ForbiddenError(
                'Client does not have correct permissions.',
            )

        try:
            queue = self._mailboxes[uid]
        except KeyError as e:
            raise BadEntityIdError(uid) from e
        try:
            return await asyncio.wait_for(queue.get(), timeout=timeout)
        except QueueShutDown:
            raise MailboxTerminatedError(uid) from None

    async def put(self, client: str | None, message: Message[Any]) -> None:
        if not self.has_permissions(client, message.dest):
            raise ForbiddenError(
                'Client does not have correct permissions.',
            )

        try:
            queue = self._mailboxes[message.dest]
        except KeyError as e:
            raise BadEntityIdError(message.dest) from e

        async with self._locks[message.dest]:
            try:
                await queue.put(message)
            except QueueShutDown:
                raise MailboxTerminatedError(message.dest) from None


async def _drain_queue(queue: AsyncQueue[Message[Any]]) -> list[Message[Any]]:
    items: list[Message[Any]] = []

    while True:
        try:
            item = queue.get_nowait()
        except (QueueShutDown, QueueEmpty):
            break
        else:
            items.append(item)
            queue.task_done()

    return items


MANAGER_KEY = AppKey('manager', _MailboxManager)


async def _create_mailbox_route(request: Request) -> Response:
    data = await request.json()
    manager: _MailboxManager = request.app[MANAGER_KEY]

    try:
        raw_mailbox_id = data['mailbox']
        mailbox_id: EntityId = TypeAdapter(EntityId).validate_json(
            raw_mailbox_id,
        )
        agent_raw = data.get('agent', None)
        agent = agent_raw.split(',') if agent_raw is not None else None
    except (KeyError, ValidationError):
        return Response(
            status=StatusCode.BAD_REQUEST.value,
            text='Missing or invalid mailbox ID',
        )

    client_id = request.headers.get('client_id', None)
    try:
        manager.create_mailbox(client_id, mailbox_id, agent)
    except ForbiddenError:
        return Response(
            status=StatusCode.FORBIDDEN.value,
            text='Incorrect permissions',
        )
    return Response(status=StatusCode.OKAY.value)


async def _terminate_route(request: Request) -> Response:
    data = await request.json()
    manager: _MailboxManager = request.app[MANAGER_KEY]

    try:
        raw_mailbox_id = data['mailbox']
        mailbox_id: EntityId = TypeAdapter(EntityId).validate_json(
            raw_mailbox_id,
        )
    except (KeyError, ValidationError):
        return Response(
            status=StatusCode.BAD_REQUEST.value,
            text='Missing or invalid mailbox ID',
        )

    client_id = request.headers.get('client_id', None)
    try:
        await manager.terminate(client_id, mailbox_id)
    except ForbiddenError:
        return Response(
            status=StatusCode.FORBIDDEN.value,
            text='Incorrect permissions',
        )
    return Response(status=StatusCode.OKAY.value)


async def _discover_route(request: Request) -> Response:
    data = await request.json()
    manager: _MailboxManager = request.app[MANAGER_KEY]

    try:
        agent = data['agent']
        allow_subclasses = data['allow_subclasses']
    except (KeyError, ValidationError):
        return Response(
            status=StatusCode.BAD_REQUEST.value,
            text='Missing or invalid arguments',
        )

    client_id = request.headers.get('client_id', None)
    agent_ids = await manager.discover(
        client_id,
        agent,
        allow_subclasses,
    )

    return json_response(
        {'agent_ids': ','.join(str(aid.uid) for aid in agent_ids)},
    )


async def _check_mailbox_route(request: Request) -> Response:
    data = await request.json()
    manager: _MailboxManager = request.app[MANAGER_KEY]

    try:
        raw_mailbox_id = data['mailbox']
        mailbox_id: EntityId = TypeAdapter(EntityId).validate_json(
            raw_mailbox_id,
        )
    except (KeyError, ValidationError):
        return Response(
            status=StatusCode.BAD_REQUEST.value,
            text='Missing or invalid mailbox ID',
        )

    client_id = request.headers.get('client_id', None)
    try:
        status = await manager.check_mailbox(client_id, mailbox_id)
    except ForbiddenError:
        return Response(
            status=StatusCode.FORBIDDEN.value,
            text='Incorrect permissions',
        )
    return json_response({'status': status.value})


async def _send_message_route(request: Request) -> Response:
    data = await request.json()
    manager: _MailboxManager = request.app[MANAGER_KEY]

    try:
        raw_message = data.get('message')
        message: Message[Any] = Message.model_validate_json(raw_message)
    except (KeyError, ValidationError):
        return Response(
            status=StatusCode.BAD_REQUEST.value,
            text='Missing or invalid message',
        )

    client_id = request.headers.get('client_id', None)
    try:
        await manager.put(client_id, message)
    except BadEntityIdError:
        return Response(
            status=StatusCode.NOT_FOUND.value,
            text='Unknown mailbox ID',
        )
    except MailboxTerminatedError:
        return Response(
            status=StatusCode.TERMINATED.value,
            text='Mailbox was closed',
        )
    except ForbiddenError:
        return Response(
            status=StatusCode.FORBIDDEN.value,
            text='Incorrect permissions',
        )
    else:
        return Response(status=StatusCode.OKAY.value)


async def _recv_message_route(request: Request) -> Response:  # noqa: PLR0911
    try:
        data = await request.json()
    except ConnectionResetError:  # pragma: no cover
        # This happens when the client cancel's it's listener task, which is
        # waiting on recv, because the client is shutting down and closing
        # its connection. In this case, we don't need to do anything
        # because the client disconnected itself. If we don't catch this
        # error, aiohttp will just log an error message each time this happens.
        return Response(status=StatusCode.NO_RESPONSE.value)

    manager: _MailboxManager = request.app[MANAGER_KEY]

    try:
        raw_mailbox_id = data['mailbox']
        mailbox_id: EntityId = TypeAdapter(EntityId).validate_json(
            raw_mailbox_id,
        )
    except (KeyError, ValidationError):
        return Response(
            status=StatusCode.BAD_REQUEST.value,
            text='Missing or invalid mailbox ID',
        )

    timeout = data.get('timeout', None)

    try:
        client_id = request.headers.get('client_id', None)
        message = await manager.get(client_id, mailbox_id, timeout=timeout)
    except BadEntityIdError:
        return Response(
            status=StatusCode.NOT_FOUND.value,
            text='Unknown mailbox ID',
        )
    except MailboxTerminatedError:
        return Response(
            status=StatusCode.TERMINATED.value,
            text='Mailbox was closed',
        )
    except ForbiddenError:
        return Response(
            status=StatusCode.FORBIDDEN.value,
            text='Incorrect permissions',
        )
    except asyncio.TimeoutError:
        return Response(
            status=StatusCode.TIMEOUT.value,
            text='Request timeout',
        )
    else:
        return json_response({'message': message.model_dump_json()})


def authenticate_factory(
    authenticator: Authenticator,
) -> Any:
    """Create an authentication middleware for a given authenticator.

    Args:
        authenticator: Used to validate client id and transform token into id.

    Returns:
        A aiohttp.web.middleware function that will only allow authenticated
            requests.
    """

    @middleware
    async def authenticate(
        request: Request,
        handler: Callable[[Request], Awaitable[Response]],
    ) -> Response:
        loop = asyncio.get_running_loop()
        try:
            # Needs to be run in executor because globus client is blocking
            client_uuid: uuid.UUID = await loop.run_in_executor(
                None,
                authenticator.authenticate_user,
                request.headers,
            )
        except ForbiddenError:
            return Response(
                status=StatusCode.FORBIDDEN.value,
                text='Token expired or revoked.',
            )
        except UnauthorizedError:
            return Response(
                status=StatusCode.UNAUTHORIZED.value,
                text='Missing required headers.',
            )

        headers = request.headers.copy()
        headers['client_id'] = str(client_uuid)

        # Handle early client-side disconnect in Issue #142
        # This is somewhat hard to reproduce in tests:
        # https://github.com/aio-libs/aiohttp/issues/6978
        if (
            request.transport is None or request.transport.is_closing()
        ):  # pragma: no cover
            return Response(status=StatusCode.NO_RESPONSE.value)

        request = request.clone(headers=headers)
        return await handler(request)

    return authenticate


def create_app(
    auth_config: ExchangeAuthConfig | None = None,
) -> Application:
    """Create a new server application."""
    middlewares = []
    if auth_config is not None:
        authenticator = get_authenticator(auth_config)
        middlewares.append(authenticate_factory(authenticator))

    manager = _MailboxManager()
    app = Application(middlewares=middlewares)
    app[MANAGER_KEY] = manager

    app.router.add_post('/mailbox', _create_mailbox_route)
    app.router.add_delete('/mailbox', _terminate_route)
    app.router.add_get('/mailbox', _check_mailbox_route)
    app.router.add_put('/message', _send_message_route)
    app.router.add_get('/message', _recv_message_route)
    app.router.add_get('/discover', _discover_route)

    return app


def _run(
    config: ExchangeServingConfig,
) -> None:
    app = create_app(config.auth)
    init_logging(config.log_level, logfile=config.log_file)
    logger = logging.getLogger('root')
    logger.info(
        'Exchange listening on %s:%s (ctrl-C to exit)',
        config.host,
        config.port,
    )

    ssl_context: ssl.SSLContext | None = None
    if config.certfile is not None:  # pragma: no cover
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        ssl_context.load_cert_chain(config.certfile, keyfile=config.keyfile)

    run_app(
        app,
        host=config.host,
        port=config.port,
        print=None,
        ssl_context=ssl_context,
    )
    logger.info('Exchange closed!')


def _main(argv: Sequence[str] | None = None) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', required=True)

    argv = sys.argv[1:] if argv is None else argv
    args = parser.parse_args(argv)

    server_config = ExchangeServingConfig.from_toml(args.config)
    _run(server_config)

    return 0



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor

import pytest

from academy.agent import action
from academy.agent import Agent
from academy.exchange import LocalExchangeFactory
from academy.handle import Handle
from academy.manager import Manager


class Coordinator(Agent):
    def __init__(
        self,
        lowerer: Handle[Lowerer],
        reverser: Handle[Reverser],
    ) -> None:
        super().__init__()
        self.lowerer = lowerer
        self.reverser = reverser

    @action
    async def process(self, text: str) -> str:
        text = await self.lowerer.lower(text)
        text = await self.reverser.reverse(text)
        return text


class Lowerer(Agent):
    @action
    async def lower(self, text: str) -> str:
        return text.lower()


class Reverser(Agent):
    @action
    async def reverse(self, text: str) -> str:
        return text[::-1]


@pytest.mark.asyncio
async def test_agent_agent_handles() -> None:
    async with await Manager.from_exchange_factory(
        factory=LocalExchangeFactory(),
        executors=ThreadPoolExecutor(),
    ) as manager:
        lowerer = await manager.launch(Lowerer)
        reverser = await manager.launch(Reverser)
        coordinator = await manager.launch(
            Coordinator,
            args=(lowerer, reverser),
        )

        text = 'DEADBEEF'
        expected = 'feebdaed'

        result = await coordinator.process(text)
        assert result == expected



from __future__ import annotations

import abc
import logging
import sys
from collections.abc import Coroutine
from typing import Any
from typing import Callable
from typing import Generic

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import TypeAlias
else:  # pragma: <3.11 cover
    from typing_extensions import TypeAlias

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    pass
else:  # pragma: <3.11 cover
    pass

from academy.agent import AgentT
from academy.exception import BadEntityIdError
from academy.exchange.client import AgentExchangeClient
from academy.exchange.client import UserExchangeClient
from academy.exchange.transport import AgentRegistration
from academy.exchange.transport import ExchangeTransportT
from academy.exchange.transport import MailboxStatus
from academy.identifier import AgentId
from academy.identifier import EntityId
from academy.identifier import UserId
from academy.message import Message
from academy.message import RequestT_co

logger = logging.getLogger(__name__)

RequestHandler: TypeAlias = Callable[
    [Message[RequestT_co]],
    Coroutine[None, None, None],
]


class ExchangeFactory(abc.ABC, Generic[ExchangeTransportT]):
    """Exchange client factory.

    An exchange factory is used to mint new exchange clients for users and
    agents, encapsulating the complexities of instantiating the underlying
    communication classes (the
    [`ExchangeTransport`][academy.exchange.transport.ExchangeTransport]).

    Warning:
        Factory implementations must be efficiently pickleable because
        factory instances are shared between user and agent processes so
        that all entities can create clients to the same exchange.
    """

    @abc.abstractmethod
    async def _create_transport(
        self,
        mailbox_id: EntityId | None = None,
        *,
        name: str | None = None,
        registration: AgentRegistration[Any] | None = None,
    ) -> ExchangeTransportT: ...

    async def create_agent_client(
        self,
        registration: AgentRegistration[AgentT],
        request_handler: RequestHandler[RequestT_co],
    ) -> AgentExchangeClient[AgentT, ExchangeTransportT]:
        """Create a new agent exchange client.

        An agent must be registered with the exchange before an exchange
        client can be created. For example:
        ```python
        factory = ExchangeFactory(...)
        user_client = factory.create_user_client()
        registration = user_client.register_agent(...)
        agent_client = factory.create_agent_client(registration, ...)
        ```

        Args:
            registration: Registration information returned by the exchange.
            request_handler: Agent request message handler.

        Returns:
            Agent exchange client.

        Raises:
            BadEntityIdError: If an agent with `registration.agent_id` is not
                already registered with the exchange.
        """
        agent_id: AgentId[AgentT] = registration.agent_id
        transport = await self._create_transport(
            mailbox_id=agent_id,
            registration=registration,
        )
        assert transport.mailbox_id == agent_id
        status = await transport.status(agent_id)
        if status != MailboxStatus.ACTIVE:
            await transport.close()
            raise BadEntityIdError(agent_id)
        return AgentExchangeClient(
            agent_id,
            transport,
            request_handler=request_handler,
        )

    async def create_user_client(
        self,
        *,
        name: str | None = None,
        start_listener: bool = True,
    ) -> UserExchangeClient[ExchangeTransportT]:
        """Create a new user in the exchange and associated client.

        Args:
            name: Display name of the client on the exchange.
            start_listener: Start a message listener thread.

        Returns:
            User exchange client.
        """
        transport = await self._create_transport(mailbox_id=None, name=name)
        user_id = transport.mailbox_id
        assert isinstance(user_id, UserId)
        return UserExchangeClient(
            user_id,
            transport,
            start_listener=start_listener,
        )



from __future__ import annotations

# Time limit on waiting for a socket connection to establish
TEST_CONNECTION_TIMEOUT = 15
# Short sleep interval used in tight loops or on timeouts that should
# immediately be triggered
TEST_SLEEP_INTERVAL = 0.001
# Time limit on calls to wait() that should return immediately
TEST_WAIT_TIMEOUT = 1



from __future__ import annotations

from testing.fixture import exchange_client
from testing.fixture import get_factory
from testing.fixture import http_exchange_factory
from testing.fixture import http_exchange_server
from testing.fixture import hybrid_exchange_factory
from testing.fixture import local_exchange_factory
from testing.fixture import manager
from testing.fixture import redis_exchange_factory
from testing.redis import mock_redis
from testing.ssl import ssl_context



from __future__ import annotations

import asyncio
import logging
from concurrent.futures import ThreadPoolExecutor

from academy.agent import action
from academy.agent import Agent
from academy.agent import loop
from academy.exchange import LocalExchangeFactory
from academy.logging import init_logging
from academy.manager import Manager

logger = logging.getLogger(__name__)


class Counter(Agent):
    count: int

    async def agent_on_startup(self) -> None:
        self.count = 0

    @loop
    async def increment(self, shutdown: asyncio.Event) -> None:
        while not shutdown.is_set():
            await asyncio.sleep(1)
            self.count += 1

    @action
    async def get_count(self) -> int:
        return self.count


async def main() -> int:
    init_logging(logging.INFO)

    async with await Manager.from_exchange_factory(
        factory=LocalExchangeFactory(),
        executors=ThreadPoolExecutor(),
    ) as manager:
        agent = await manager.launch(Counter)

        logger.info('Waiting 2s for agent loops to execute...')
        await asyncio.sleep(2)

        count = await agent.get_count()
        assert count >= 1
        logger.info('Agent loop executed %s time(s)', count)

    return 0


if __name__ == '__main__':
    raise SystemExit(asyncio.run(main()))



from __future__ import annotations

import os
import pathlib
import uuid
from collections.abc import Generator
from unittest import mock

import globus_sdk
import pytest
from globus_sdk.globus_app import ClientApp
from globus_sdk.globus_app import GlobusAppConfig
from globus_sdk.globus_app import UserApp
from globus_sdk.tokenstorage import MemoryTokenStorage
from globus_sdk.tokenstorage import TokenValidationError

from academy.exchange.cloud.login import ACADEMY_GLOBUS_CLIENT_ID_ENV_NAME
from academy.exchange.cloud.login import ACADEMY_GLOBUS_CLIENT_SECRET_ENV_NAME
from academy.exchange.cloud.login import get_auth_headers
from academy.exchange.cloud.login import get_client_app
from academy.exchange.cloud.login import get_client_credentials_from_env
from academy.exchange.cloud.login import get_globus_app
from academy.exchange.cloud.login import get_token_storage
from academy.exchange.cloud.login import get_user_app
from academy.exchange.cloud.login import is_client_login


@pytest.fixture
def mock_env_credentials() -> Generator[tuple[str, str], None, None]:
    client_uuid = str(uuid.uuid4())
    client_secret = 'secret'

    env = {
        ACADEMY_GLOBUS_CLIENT_ID_ENV_NAME: client_uuid,
        ACADEMY_GLOBUS_CLIENT_SECRET_ENV_NAME: client_secret,
    }
    with mock.patch.dict(os.environ, env):
        yield client_uuid, client_secret


def test_get_token_storage(tmp_path: pathlib.Path):
    filepath = tmp_path / 'tokens.db'
    storage = get_token_storage(filepath)
    assert filepath.is_file()
    storage.close()


def test_get_token_storage_academy_default(tmp_path: pathlib.Path):
    env = {
        'ACADEMY_HOME': str(tmp_path),
    }
    with mock.patch.dict(os.environ, env):
        filepath = tmp_path / 'storage.db'
        storage = get_token_storage()
        assert filepath.is_file()
        storage.close()


def test_get_confidential_app_auth_client_from_env(
    mock_env_credentials,
) -> None:
    found_id, found_secret = get_client_credentials_from_env()
    assert found_id == mock_env_credentials[0]
    assert found_secret == mock_env_credentials[1]


@mock.patch('academy.exchange.cloud.login.get_token_storage')
def test_get_globus_app_client_login(
    mock_storage,
    mock_env_credentials,
) -> None:
    mock_storage.return_value = MemoryTokenStorage()
    globus_app = get_globus_app()
    assert isinstance(globus_app, ClientApp)


@mock.patch('academy.exchange.cloud.login.get_token_storage')
def test_get_globus_app_not_client_login(mock_storage) -> None:
    mock_storage.return_value = MemoryTokenStorage()
    globus_app = get_globus_app()
    assert isinstance(globus_app, UserApp)


@mock.patch('academy.exchange.cloud.login.get_token_storage')
def test_get_client_app_from_env(mock_storage, mock_env_credentials) -> None:
    mock_storage.return_value = MemoryTokenStorage()
    globus_app = get_client_app()
    assert isinstance(globus_app, ClientApp)


@mock.patch('academy.exchange.cloud.login.get_token_storage')
def test_get_client_app_custom(mock_storage) -> None:
    mock_storage.return_value = MemoryTokenStorage()
    globus_app = get_client_app(str(uuid.uuid4()), '<secret>')
    assert isinstance(globus_app, ClientApp)


@mock.patch('academy.exchange.cloud.login.get_token_storage')
def test_get_user_app(mock_storage) -> None:
    mock_storage.return_value = MemoryTokenStorage()
    globus_app = get_user_app()
    assert isinstance(globus_app, UserApp)


def test_is_client_login() -> None:
    with mock.patch.dict(os.environ, {}):
        assert not is_client_login()

    env = {ACADEMY_GLOBUS_CLIENT_ID_ENV_NAME: str(uuid.uuid4())}
    with mock.patch.dict(os.environ, env):
        assert not is_client_login()

    env[ACADEMY_GLOBUS_CLIENT_SECRET_ENV_NAME] = 'secret'
    with mock.patch.dict(os.environ, env):
        assert is_client_login()


def test_get_auth_headers_none() -> None:
    assert get_auth_headers(None) == {}


@pytest.fixture
def globus_app() -> UserApp:
    config = GlobusAppConfig(token_storage=MemoryTokenStorage())
    mock_client = mock.Mock(
        spec=globus_sdk.NativeAppAuthClient,
        client_id='mock-client_id',
        base_url='https://auth.globus.org',
        environment='production',
    )
    return UserApp(
        app_name='test-app',
        login_client=mock_client,
        config=config,
    )


def test_get_auth_headers_globus(globus_app) -> None:
    mock_authorizer = mock.MagicMock()
    header = 'Bearer <TOKEN>'

    with (
        mock.patch(
            'academy.exchange.cloud.login.get_globus_app',
            return_value=globus_app,
        ),
        mock.patch.object(
            globus_app,
            'get_authorizer',
            return_value=mock_authorizer,
        ),
        mock.patch.object(
            mock_authorizer,
            'get_authorization_header',
            return_value=header,
        ),
    ):
        assert get_auth_headers('globus')['Authorization'] == header


def test_get_auth_headers_globus_missing(globus_app) -> None:
    with (
        mock.patch(
            'academy.exchange.cloud.login.get_globus_app',
            return_value=globus_app,
        ),
        mock.patch.object(
            globus_app,
            'get_authorizer',
            side_effect=TokenValidationError(),
        ),
        pytest.raises(
            SystemExit,
        ),
    ):
        assert get_auth_headers('globus')



from __future__ import annotations

import asyncio
import contextlib
import dataclasses
import logging
import sys
import warnings
from collections.abc import Iterable
from collections.abc import MutableMapping
from concurrent.futures import Executor
from types import TracebackType
from typing import Any
from typing import Generic

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import Self
else:  # pragma: <3.11 cover
    from typing_extensions import Self

from academy.agent import AgentT
from academy.exception import AgentTerminatedError
from academy.exception import BadEntityIdError
from academy.exception import raise_exceptions
from academy.exchange import ExchangeFactory
from academy.exchange import UserExchangeClient
from academy.exchange.transport import AgentRegistration
from academy.exchange.transport import ExchangeTransportT
from academy.handle import exchange_context
from academy.handle import Handle
from academy.identifier import AgentId
from academy.identifier import UserId
from academy.runtime import Runtime
from academy.runtime import RuntimeConfig
from academy.serialize import NoPickleMixin

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class _RunSpec(Generic[AgentT, ExchangeTransportT]):
    agent: AgentT | type[AgentT]
    config: RuntimeConfig
    exchange_factory: ExchangeFactory[ExchangeTransportT]
    registration: AgentRegistration[AgentT]
    agent_args: tuple[Any, ...]
    agent_kwargs: dict[str, Any]


async def _run_agent_on_worker_async(
    spec: _RunSpec[AgentT, ExchangeTransportT],
) -> None:
    try:
        if isinstance(spec.agent, type):
            agent = spec.agent(*spec.agent_args, **spec.agent_kwargs)
        else:
            agent = spec.agent

        async with Runtime(
            agent,
            config=spec.config,
            exchange_factory=spec.exchange_factory,
            registration=spec.registration,
        ) as runtime:
            await runtime.wait_shutdown()
    except:
        logger.exception(
            'Failure running agent on worker (%s)',
            spec.registration.agent_id,
        )
        raise


def _run_agent_on_worker(
    spec: _RunSpec[AgentT, ExchangeTransportT],
) -> None:
    asyncio.run(_run_agent_on_worker_async(spec))


@dataclasses.dataclass
class _ACB(Generic[AgentT]):
    # Agent Control Block
    agent_id: AgentId[AgentT]
    executor: str
    task: asyncio.Task[None]


class Manager(Generic[ExchangeTransportT], NoPickleMixin):
    """Launch and manage running agents.

    A manager is used to launch agents using one or more
    [`Executors`][concurrent.futures.Executor] and interact with/manage those
    agents.

    Tip:
        This class can be used as a context manager. Upon exiting the context,
        running agents will be shutdown, any agent handles created by the
        manager will be closed, and the executors will be shutdown.

    Tip:
        When using
        [`ProcessPoolExecutors`][concurrent.futures.ProcessPoolExecutor],
        use the `initializer` argument to configure logging in the worker
        processes that will execute agents.

    Note:
        The manager takes ownership of the exchange client and executors,
        meaning the manager will clean up those resources when the manager
        is closed.

    Args:
        exchange_client: Exchange client.
        executors: An executor instance or mapping of names to executors to
            use to run agents. If a single executor is provided, it is set
            as the default executor with name `'default'`, overriding any
            value of `default_executor`.
        default_executor: Specify the name of the default executor to use
            when not specified in `launch()`.
        max_retries: Maximum number of times to retry running an agent
            if it exits with an error.

    Raises:
        ValueError: If `default_executor` is specified but does not exist
            in `executors`.
    """

    def __init__(
        self,
        exchange_client: UserExchangeClient[ExchangeTransportT],
        executors: Executor | MutableMapping[str, Executor],
        *,
        default_executor: str | None = None,
        max_retries: int = 0,
    ) -> None:
        if isinstance(executors, Executor):
            executors = {'default': executors}
            default_executor = 'default'

        if default_executor is not None and default_executor not in executors:
            raise ValueError(
                f'No executor named "{default_executor}" was provided to '
                'use as the default.',
            )

        self._exchange_client = exchange_client
        self._exchange_factory = exchange_client.factory()
        self._user_id = self._exchange_client.client_id
        self._executors = executors
        self._default_executor = default_executor
        self._max_retries = max_retries

        self._handles: dict[AgentId[Any], Handle[Any]] = {}
        self._acbs: dict[AgentId[Any], _ACB[Any]] = {}

        logger.info('Initialized manager (%s)', self.user_id)

    async def __aenter__(self) -> Self:
        self.exchange_context_token = exchange_context.set(
            self.exchange_client,
        )
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        await self.close()
        exchange_context.reset(self.exchange_context_token)

    def __repr__(self) -> str:
        executors_repr = ', '.join(
            f'{k}: {v!r}' for k, v in self._executors.items()
        )
        return (
            f'{type(self).__name__}'
            f'(exchange={self._exchange_client!r}, '
            f'executors={{{executors_repr}}})'
        )

    def __str__(self) -> str:
        return (
            f'{type(self).__name__}<{self.user_id}, {self._exchange_client}>'
        )

    @classmethod
    async def from_exchange_factory(
        cls,
        factory: ExchangeFactory[ExchangeTransportT],
        executors: Executor | MutableMapping[str, Executor],
        *,
        default_executor: str | None = None,
        max_retries: int = 0,
    ) -> Self:
        """Instantiate a new exchange client and manager from a factory."""
        client = await factory.create_user_client()
        return cls(
            client,
            executors,
            default_executor=default_executor,
            max_retries=max_retries,
        )

    @property
    def exchange_client(self) -> UserExchangeClient[ExchangeTransportT]:
        """User client for the exchange."""
        return self._exchange_client

    @property
    def exchange_factory(self) -> ExchangeFactory[ExchangeTransportT]:
        """Client factory for the exchange."""
        return self._exchange_factory

    @property
    def user_id(self) -> UserId:
        """Exchange client user ID of this manager."""
        return self._user_id

    async def close(self) -> None:
        """Shutdown the manager and cleanup resources.

        1. Request all running agents to shut down.
        1. Wait for all running agents to shut down.
        1. Close the exchange client.
        1. Shutdown the executors.
        1. Raise an exceptions returned by agents.

        Raises:
            Exception: Any exceptions raised by agents.
        """
        for acb in self._acbs.values():
            if not acb.task.done():
                handle = self.get_handle(acb.agent_id)
                with contextlib.suppress(AgentTerminatedError):
                    await handle.shutdown()
        logger.debug('Requested shutdown from all agents')

        for acb in self._acbs.values():
            await acb.task
        logger.debug('All agents have completed')

        await self.exchange_client.close()
        for executor in self._executors.values():
            executor.shutdown(wait=True, cancel_futures=True)

        exceptions = (acb.task.exception() for acb in self._acbs.values())
        exceptions_only = tuple(e for e in exceptions if e is not None)
        raise_exceptions(
            exceptions_only,
            message='Caught failures in agent while shutting down.',
        )

        logger.info('Closed manager (%s)', self.user_id)

    def add_executor(self, name: str, executor: Executor) -> Self:
        """Add an executor to the manager.

        Note:
            It is not possible to remove an executor as this could create
            complications if an agent is already running in that executor.

        Args:
            name: Name of the executor used when launching agents.
            executor: Executor instance.

        Returns:
            Self for chaining.

        Raises:
            ValueError: If an executor with `name` already exists.
        """
        if name in self._executors:
            raise ValueError(f'Executor named "{name}" already exists.')
        self._executors[name] = executor
        return self

    def set_default_executor(self, name: str | None) -> Self:
        """Set the default executor by name.

        Args:
            name: Name of the executor to use as default. If `None`, no
                default executor is set and all calls to `launch()` must
                specify the executor.

        Returns:
            Self for chaining.

        Raises:
            ValueError: If no executor with `name` exists.
        """
        if name not in self._executors:
            raise ValueError(f'An executor named "{name}" does not exist.')
        self._default_executor = name
        return self

    async def _run_agent_in_executor(
        self,
        executor: Executor,
        spec: _RunSpec[AgentT, ExchangeTransportT],
    ) -> None:
        agent_id = spec.registration.agent_id
        original_config = spec.config
        loop = asyncio.get_running_loop()
        run_count = 0
        retries = self._max_retries

        while True:
            run_count += 1
            if retries > 0:
                retries -= 1
                # Override this configuration for the case where the agent
                # fails and we will be restarting it.
                spec.config = dataclasses.replace(
                    original_config,
                    terminate_on_error=False,
                )
            else:
                # Otherwise, keep the original config.
                spec.config = original_config

            logger.debug(
                'Launching agent (attempt: %s; retries: %s; %s; %s)',
                run_count,
                retries,
                agent_id,
                spec.agent,
            )

            try:
                await loop.run_in_executor(
                    executor,
                    _run_agent_on_worker,
                    spec,
                )
            except asyncio.CancelledError:  # pragma: no cover
                logger.warning('Cancelled %s task', agent_id)
                raise
            except Exception:
                if retries == 0:
                    logger.exception('Received exception from %s', agent_id)
                    raise
                else:
                    logger.exception(
                        'Restarting %s due to exception',
                        agent_id,
                    )
            else:
                logger.debug('Completed %s task', agent_id)
                break

    async def launch(  # noqa: PLR0913
        self,
        agent: AgentT | type[AgentT],
        *,
        args: tuple[Any, ...] | None = None,
        kwargs: dict[str, Any] | None = None,
        config: RuntimeConfig | None = None,
        executor: str | None = None,
        name: str | None = None,
        registration: AgentRegistration[AgentT] | None = None,
    ) -> Handle[AgentT]:
        """Launch a new agent with a specified agent.

        Args:
            agent: Agent instance the agent will implement or the
                agent type that will be initialized on the worker using
                `args` and `kwargs`.
            args: Positional arguments used to initialize the agent.
                Ignored if `agent` is already an instance.
            kwargs: Keyword arguments used to initialize the agent.
                Ignored if `agent` is already an instance.
            config: Agent run configuration.
            executor: Name of the executor instance to use. If `None`, uses
                the default executor, if specified, otherwise raises an error.
            name: Readable name of the agent used when registering a new agent.
            registration: If `None`, a new agent will be registered with
                the exchange.

        Returns:
            Handle (client bound) used to interact with the agent.

        Raises:
            RuntimeError: If `registration` is provided and an agent with
                that ID has already been executed.
            ValueError: If no default executor is set and `executor` is not
                specified.
        """
        if self._default_executor is None and executor is None:
            raise ValueError(
                'Must specify the executor when no default is set.',
            )
        executor = executor if executor is not None else self._default_executor
        assert executor is not None
        executor_instance = self._executors[executor]

        if registration is None:
            agent_type = agent if isinstance(agent, type) else type(agent)
            registration = await self.register_agent(agent_type, name=name)
        elif registration.agent_id in self._acbs:
            raise RuntimeError(
                f'{registration.agent_id} has already been executed.',
            )

        agent_id = registration.agent_id

        spec = _RunSpec(
            agent=agent,
            config=RuntimeConfig() if config is None else config,
            exchange_factory=self.exchange_factory,
            registration=registration,
            agent_args=() if args is None else args,
            agent_kwargs={} if kwargs is None else kwargs,
        )

        task = asyncio.create_task(
            self._run_agent_in_executor(executor_instance, spec),
            name=f'manager-run-{agent_id}',
        )

        acb = _ACB(agent_id=agent_id, executor=executor, task=task)
        self._acbs[agent_id] = acb
        handle = self.get_handle(agent_id)
        logger.info('Launched agent (%s; %s)', agent_id, agent)
        self._warn_executor_overloaded(executor_instance, executor)
        return handle

    def get_handle(
        self,
        agent: AgentId[AgentT] | AgentRegistration[AgentT],
    ) -> Handle[AgentT]:
        """Create a new handle to an agent.

        A handle acts like a reference to a remote agent, enabling a user
        to manage the agent or asynchronously invoke actions.

        Args:
            agent: Agent ID or registration indicating the agent to create
                a handle to. The agent must be registered with the same
                exchange that this manager is a client of.

        Returns:
            Handle to the agent.
        """
        agent_id = agent if isinstance(agent, AgentId) else agent.agent_id
        handle = self._handles.get(agent_id, None)
        if handle is not None:
            return handle
        handle = Handle(agent_id)
        self._handles[agent_id] = handle
        return handle

    async def register_agent(
        self,
        agent: type[AgentT],
        *,
        name: str | None = None,
    ) -> AgentRegistration[AgentT]:
        """Register a new agent with the exchange.

        Args:
            agent: Agent type of the agent.
            name: Optional display name for the agent.

        Returns:
            Agent registration info that can be passed to
            [`launch()`][academy.manager.Manager.launch].
        """
        return await self.exchange_client.register_agent(agent, name=name)

    def running(self) -> set[AgentId[Any]]:
        """Get a set of IDs of all running agents.

        Returns:
            Set of agent IDs corresponding to all agents launched by this \
            manager that have not completed yet.
        """
        running: set[AgentId[Any]] = set()
        for acb in self._acbs.values():
            if not acb.task.done():
                running.add(acb.agent_id)
        return running

    async def shutdown(
        self,
        agent: AgentId[Any] | Handle[Any],
        *,
        blocking: bool = True,
        raise_error: bool = True,
        terminate: bool | None = None,
        timeout: float | None = None,
    ) -> None:
        """Shutdown a launched agent.

        Args:
            agent: ID or handle to the launched agent.
            blocking: Wait for the agent to exit before returning.
            raise_error: Raise the error returned by the agent if
                `blocking=True`.
            terminate: Override the termination behavior of the agent defined
                in the [`RuntimeConfig`][academy.runtime.RuntimeConfig].
            timeout: Optional timeout is seconds when `blocking=True`.

        Raises:
            BadEntityIdError: If an agent with `agent_id` was not
                launched by this manager.
            TimeoutError: If `timeout` was exceeded while blocking for agent.
        """
        agent_id = agent.agent_id if isinstance(agent, Handle) else agent

        if agent_id not in self._acbs:
            raise BadEntityIdError(agent_id) from None
        if self._acbs[agent_id].task.done():
            return

        handle = self.get_handle(agent_id)
        with contextlib.suppress(AgentTerminatedError):
            await handle.shutdown(terminate=terminate)

        if blocking:
            await self.wait(
                {agent_id},
                raise_error=raise_error,
                timeout=timeout,
            )

    async def wait(
        self,
        agents: Iterable[AgentId[Any] | Handle[Any]],
        *,
        raise_error: bool = False,
        return_when: str = asyncio.ALL_COMPLETED,
        timeout: float | None = None,
    ) -> None:
        """Wait for launched agents to complete.

        Note:
            Calling `wait()` is only valid after `launch()` has succeeded.

        Args:
            agents: An iterable of agent IDs or handles to wait on.
            raise_error: Raise errors returned by completed agents.
            return_when: Indicate when this function should return. The
                same as [`asyncio.wait()`][asyncio.wait].
            timeout: Optional timeout in seconds to wait for agents.

        Raises:
            BadEntityIdError: If an agent was not launched by this manager.
            TimeoutError: If `timeout` was exceeded while waiting for agents.
            Exception: Any exception raised by an agent that completed due
                to a failure and `raise_error=True` is set.
        """
        agent_ids = {
            agent if isinstance(agent, AgentId) else agent.agent_id
            for agent in agents
        }

        if len(agent_ids) == 0:
            return

        agent_tasks: list[asyncio.Task[None]] = []
        for agent_id in agent_ids:
            try:
                agent_tasks.append(self._acbs[agent_id].task)
            except KeyError:
                raise BadEntityIdError(agent_id) from None

        done, pending = await asyncio.wait(
            agent_tasks,
            return_when=return_when,
            timeout=timeout,
        )

        if len(done) == 0:
            raise TimeoutError(
                f'No agents completed within {timeout} seconds: '
                f'{len(pending)} pending agent(s).',
            )
        elif return_when == asyncio.ALL_COMPLETED and len(pending) > 0:
            raise TimeoutError(
                f'Not all agents completed within {timeout} seconds: '
                f'{len(pending)} pending agent(s).',
            )

        if raise_error:
            exceptions = (task.exception() for task in agent_tasks)
            exceptions_only = tuple(e for e in exceptions if e is not None)
            raise_exceptions(
                exceptions_only,
                message='Waited agents raised the following exceptions.',
            )

    def _warn_executor_overloaded(
        self,
        executor: Executor,
        name: str,
    ) -> None:
        max_workers = _infer_max_workers(executor)
        if max_workers is None:  # pragma: no cover
            # If the user provided a third-party executor that we don't
            # know how to get the number of workers for just return.
            return
        running_agents = len(self.running())
        if running_agents > max_workers:
            warnings.warn(
                f'Executor overload: submitted agents exceeds worker count '
                f'(executor: {name}, running agents: {running_agents}, max '
                f'workers: {max_workers})',
                RuntimeWarning,
                # This stacklevel ensures the user see the line that called
                # Manager.launch(), rather than seeing this line or the
                # call to _warn_executor_overloaded
                stacklevel=3,
            )


def _infer_max_workers(executor: Executor) -> int | None:  # pragma: no cover
    """Infer the maximum workers of the executor.

    The [`Executor`][concurrent.futures.Executor] specification does not
    provide a standard mechanism to get the maximum number of workers.
    """
    if hasattr(executor, '_max_workers'):
        # ProcessPoolExecutor and ThreadPoolExecutor
        return executor._max_workers
    elif hasattr(executor, 'scheduler_info') and callable(
        executor.scheduler_info,
    ):
        # Dask distributed client
        try:
            return executor.scheduler_info()['workers']
        except KeyError:
            return None
    else:
        return None



from __future__ import annotations

import asyncio
import contextlib


async def wait_event_async(
    *events: asyncio.Event,
    timeout: float | None = None,
) -> asyncio.Event:
    """Wait for the first async event to be set.

    Args:
        events: One or more events to wait on.
        timeout: Maximum number of seconds to wait for an event to finish.

    Returns:
        The first event to finish.

    Raises:
        TimeoutError: If no event finished within `timeout` seconds.
    """
    tasks = {
        asyncio.create_task(event.wait(), name=f'or-event-waiter-{i}'): event
        for i, event in enumerate(events)
    }
    done, pending = await asyncio.wait(
        tasks.keys(),
        timeout=timeout,
        return_when=asyncio.FIRST_COMPLETED,
    )
    for task in pending:
        task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await task

    if len(done) == 0:
        raise TimeoutError(f'No events were set within {timeout} seconds.')

    return tasks[done.pop()]



from __future__ import annotations

import asyncio
import functools
import inspect
import logging
import sys
import warnings
from collections.abc import Coroutine
from collections.abc import Generator
from datetime import timedelta
from typing import Any
from typing import Callable
from typing import Generic
from typing import Literal
from typing import overload
from typing import Protocol
from typing import TYPE_CHECKING
from typing import TypeVar

if sys.version_info >= (3, 10):  # pragma: >=3.10 cover
    from typing import ParamSpec
    from typing import TypeAlias
else:  # pragma: <3.10 cover
    from typing_extensions import ParamSpec
    from typing_extensions import TypeAlias

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import Self
else:  # pragma: <3.11 cover
    from typing_extensions import Self

from academy.event import wait_event_async
from academy.exception import AgentNotInitializedError
from academy.handle import Handle

if TYPE_CHECKING:
    from academy.context import AgentContext
    from academy.exchange import AgentExchangeClient
    from academy.identifier import AgentId

AgentT = TypeVar('AgentT', bound='Agent')
"""Type variable bound to [`Agent`][academy.agent.Agent]."""

P = ParamSpec('P')
R = TypeVar('R')
R_co = TypeVar('R_co', covariant=True)
ActionMethod: TypeAlias = Callable[P, Coroutine[None, None, R]]
LoopMethod: TypeAlias = Callable[
    [AgentT, asyncio.Event],
    Coroutine[None, None, None],
]

logger = logging.getLogger(__name__)


class Agent:
    """Agent base class.

    An agent is composed of three parts:

    1. The [`agent_on_startup()`][academy.agent.Agent.agent_on_startup] and
       [`agent_on_shutdown()`][academy.agent.Agent.agent_on_shutdown] methods
       define callbacks that are invoked once at the start and end of an
       agent's execution, respectively. The methods should be used to
       initialize and cleanup stateful resources. Resource initialization
       should not be performed in `__init__`.
    1. Action methods annotated with [`@action`][academy.agent.action]
       are methods that other agents can invoke on this agent. An agent
       may also call it's own action methods as normal methods.
    1. Control loop methods annotated with [`@loop`][academy.agent.loop]
       are executed in separate threads when the agent is executed.

    The [`Runtime`][academy.runtime.Runtime] is used to execute an agent
    definition.

    Warning:
        This class cannot be instantiated directly and must be subclassed.
    """

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:  # noqa: D102
        if cls is Agent:
            raise TypeError(
                f'The {cls.__name__} type cannot be instantiated directly '
                'and must be subclassed.',
            )
        return super().__new__(cls)

    def __init__(self) -> None:
        self.__agent_context: AgentContext[Self] | None = None
        self.__agent_run_sync_semaphore: asyncio.Semaphore | None = None

    def __repr__(self) -> str:
        return f'{type(self).__name__}()'

    def __str__(self) -> str:
        return f'Agent<{type(self).__name__}>'

    def _agent_set_context(self, context: AgentContext[Self]) -> None:
        self.__agent_context = context

    @property
    def agent_context(self) -> AgentContext[Self]:
        """Agent runtime context.

        Raises:
            AgentNotInitializedError: If the agent runtime implementing
                this agent has not been started.
        """
        if (
            # Check _Agent__agent_context rather than __agent_context
            # because of Python's name mangling
            not hasattr(self, '_Agent__agent_context')
            or self.__agent_context is None
        ):
            raise AgentNotInitializedError
        return self.__agent_context

    @property
    def agent_id(self) -> AgentId[Self]:
        """Agent Id.

        Raises:
            AgentNotInitializedError: If the agent runtime implementing
                this agent has not been started.
        """
        return self.agent_context.agent_id

    @property
    def agent_exchange_client(self) -> AgentExchangeClient[Self, Any]:
        """Agent exchange client.

        Raises:
            AgentNotInitializedError: If the agent runtime implementing
                this agent has not been started.
        """
        return self.agent_context.exchange_client

    def _agent_attributes(self) -> Generator[tuple[str, Any]]:
        """Returns a generator that yields attributes of the agent.

        The following attributes are ignored:

        * Attributes defined on the parent/base Agent class
        * @property methods of the derived type
        """
        all_attrs = set(dir(self))
        base_attrs = set(dir(Agent))
        derived_attrs = all_attrs - base_attrs

        for name in derived_attrs:
            if name.startswith('_Agent__'):
                # Skip '__'-prefixed attributes of Agent whose names
                # were mangled in the derived type
                continue
            if isinstance(getattr(type(self), name, None), property):
                # Skip checking properties defined on the derived type.
                continue
            attr = getattr(self, name)
            yield name, attr

    def _agent_actions(self) -> dict[str, Action[Any, Any]]:
        """Get methods of this agent type that are decorated as actions.

        Returns:
            Dictionary mapping method names to action methods.
        """
        actions: dict[str, Action[Any, Any]] = {}
        for name, attr in self._agent_attributes():
            if _is_agent_method_type(attr, 'action'):
                actions[name] = attr
        return actions

    def _agent_loops(self) -> dict[str, ControlLoop]:
        """Get methods of this agent type that are decorated as loops.

        Returns:
            Dictionary mapping method names to loop methods.
        """
        loops: dict[str, ControlLoop] = {}
        for name, attr in self._agent_attributes():
            if _is_agent_method_type(attr, 'loop'):
                loops[name] = attr
        return loops

    @classmethod
    def _agent_mro(cls) -> tuple[str, ...]:
        """Get the method resolution order of the agent.

        Example:
            ```python
            >>> from academy.agent import Agent
            >>>
            >>> class A(Agent): ...
            >>> class B(Agent): ...
            >>> class C(A): ...
            >>> class D(A, B): ...
            >>>
            >>> A._agent_mro()
            ('__main__.A',)
            >>> B._agent_mro()
            ('__main__.B',)
            >>> C._agent_mro()
            ('__main__.C', '__main__.A')
            >>> D._agent_mro()
            ('__main__.D', '__main__.A', '__main__.B')
            ```

        Returns:
            Tuple of fully-qualified paths of types in the MRO of this \
            agent type, not including the base \
            [`Agent`][academy.agent.Agent] or [`object`][object].
        """
        mro = cls.mro()
        base_index = mro.index(Agent)
        mro = mro[:base_index]
        return tuple(f'{t.__module__}.{t.__qualname__}' for t in mro)

    async def agent_on_startup(self) -> None:
        """Callback invoked at the end of an agent's startup sequence.

        Control loops will not start and action requests will not be
        processed until after this callback completes. Thus, it is safe to
        initialize resources in this callback that are needed by actions or
        loops.

        See
        [`Runtime.run_until_complete()`][academy.runtime.Runtime.run_until_complete]
        for more details on the startup sequence.
        """
        pass

    async def agent_on_shutdown(self) -> None:
        """Callback invoked at the beginning of an agent's shutdown sequence.

        See
        [`Runtime.run_until_complete()`][academy.runtime.Runtime.run_until_complete]
        for more details on the shutdown sequence.
        """
        pass

    async def agent_run_sync(
        self,
        function: Callable[P, R],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> R:
        """Run a blocking function in separate thread.

        Example:
            ```python
            import time
            from academy.agent import Agent, action

            class Example(Agent):
                def blocking_call(self, value: int) -> int:
                    time.sleep(10)
                    return value

                @action
                async def non_blocking_call(self, value: int) -> int:
                    result = await self.agent_run_sync(self.blocking_call, value)
                    ...
                    return result
            ```

        Note:
            The max concurrency of the executor is configured in the
            [`RuntimeConfig`][academy.runtime.RuntimeConfig]. If all
            executor workers are busy the function will be queued and a
            warning will be logged.

        Warning:
           This function does not support cancellation. For example, if you
           wrap this call in [`asyncio.wait_for()`][asyncio.wait_for] and a
           timeout occurs, the task wrapping the coroutine will be cancelled
           but the blocking function will continue running in its thread until
           completion.

        Args:
            function: The blocking function to run.
            *args: Positional arguments for the function.
            **kwargs: Keyword arguments for the function.

        Returns:
            The result of the function call.

        Raises:
            AgentNotInitializedError: If the agent runtime has not been
                started.
            Exception: Any exception raised by the function.
        """  # noqa: E501
        executor = self.agent_context.executor

        wrapped = functools.partial(function, *args, **kwargs)
        loop = asyncio.get_running_loop()

        if self.__agent_run_sync_semaphore is None:
            max_workers = executor._max_workers
            self.__agent_run_sync_semaphore = asyncio.Semaphore(max_workers)

        acquired = self.__agent_run_sync_semaphore.locked()
        if acquired:
            logger.warning(
                f'Thread-pool executor for {self.agent_id} is overloaded, '
                f'sync function "{function.__name__}" is waiting for a '
                'worker',
            )

        async with self.__agent_run_sync_semaphore:
            return await loop.run_in_executor(executor, wrapped)

    def agent_shutdown(self) -> None:
        """Request the agent to shutdown.

        Raises:
            AgentNotInitializedError: If the agent runtime implementing
                this agent has not been started.
        """
        self.agent_context.shutdown_event.set()


class Action(Generic[P, R_co], Protocol):
    """Action method protocol."""

    _agent_method_type: Literal['action']
    _action_method_context: bool

    async def __call__(self, *arg: P.args, **kwargs: P.kwargs) -> R_co:
        """Expected signature of methods decorated as an action.

        In general, action methods can implement any signature.
        """
        ...


class ControlLoop(Protocol):
    """Control loop method protocol."""

    _agent_method_type: Literal['loop']

    async def __call__(self, shutdown: asyncio.Event) -> None:
        """Expected signature of methods decorated as a control loop.

        Args:
            shutdown: Event indicating that the agent has been instructed to
                shutdown and all control loops should exit.

        Returns:
            Control loops should not return anything.
        """
        ...


@functools.lru_cache(maxsize=1)
def _get_handle_protected_methods() -> tuple[str, ...]:
    methods: list[str] = []
    for name, value in inspect.getmembers(Handle):
        # Only include functions defined on Handle, not inherited ones
        if inspect.isfunction(value) and name in Handle.__dict__:
            methods.append(name)
    return tuple(methods)


@overload
def action(method: ActionMethod[P, R]) -> ActionMethod[P, R]: ...


@overload
def action(
    *,
    allow_protected_name: bool = False,
    context: bool = False,
) -> Callable[[ActionMethod[P, R]], ActionMethod[P, R]]: ...


def action(
    method: ActionMethod[P, R] | None = None,
    *,
    allow_protected_name: bool = False,
    context: bool = False,
) -> ActionMethod[P, R] | Callable[[ActionMethod[P, R]], ActionMethod[P, R]]:
    """Decorator that annotates a method of a agent as an action.

    Marking a method of a agent as an action makes the method available
    to other agents. I.e., peers within a multi-agent system can only invoke
    methods marked as actions on each other. This enables agents to
    define "private" methods.

    Example:
        ```python
        from academy.agent import Agent, action
        from academy.context import ActionContext

        class Example(Agent):
            @action
            async def perform(self) -> ...:
                ...

            @action(context=True)
            async def perform_with_ctx(self, *, context: ActionContext) -> ...:
                ...
        ```

    Warning:
        A warning will be emitted if the decorated method's name clashed
        with a method of [`Handle`][academy.handle.Handle] because it would
        not be possible to invoke this action remotely via attribute
        lookup on a handle. This warning can be suppressed with
        `allow_protected_name=True`, and the action must be invoked via
        [`Handle.action()`][academy.handle.Handle.action].

    Args:
        method: Method to decorate as an action.
        allow_protected_name: Allow decorating a method as an action when
            the name of the method clashes with a protected method name of
            [`Handle`][academy.handle.Handle]. This flag silences the
            emitted warning.
        context: Specify that the action method expects a context argument.
            The `context` will be provided at runtime as a keyword argument.

    Raises:
        TypeError: If `context=True` and the method does not have a parameter
            named `context` or if `context` is a positional only argument.
    """

    def decorator(method_: ActionMethod[P, R]) -> ActionMethod[P, R]:
        if (
            not allow_protected_name
            and method_.__name__ in _get_handle_protected_methods()
        ):
            warnings.warn(
                f'The name of the decorated method is "{method_.__name__}" '
                'which clashes with a protected method of Handle. '
                'Rename the decorated method to avoid ambiguity when remotely '
                'invoking it via a handle.',
                UserWarning,
                stacklevel=3,
            )
        # Typing the requirement that if context=True then params P should
        # contain a keyword argument named "context" is not easily annotated
        # for mypy so instead we check at runtime.
        if context:
            sig = inspect.signature(method_)
            if 'context' not in sig.parameters:
                raise TypeError(
                    f'Action method "{method_.__name__}" must accept a '
                    '"context" keyword argument when used with '
                    '@action(context=True).',
                )
            if (
                sig.parameters['context'].kind
                != inspect.Parameter.KEYWORD_ONLY
            ):
                raise TypeError(
                    'The "context" argument to action method '
                    f'"{method_.__name__}" must be a keyword only argument.',
                )

        method_._agent_method_type = 'action'  # type: ignore[attr-defined]
        method_._action_method_context = context  # type: ignore[attr-defined]
        return method_

    if method is None:
        return decorator
    else:
        return decorator(method)


def loop(method: LoopMethod[AgentT]) -> LoopMethod[AgentT]:
    """Decorator that annotates a method of a agent as a control loop.

    Control loop methods of a agent are run as threads when an agent
    starts. A control loop can run for a well-defined period of time or
    indefinitely, provided the control loop exits when the `shutdown`
    event, passed as a parameter to all control loop methods, is set.

    Example:
        ```python
        import asyncio
        from academy.agent import Agent, loop

        class Example(Agent):
            @loop
            async def listen(self, shutdown: asyncio.Event) -> None:
                while not shutdown.is_set():
                    ...
        ```

    Raises:
        TypeError: if the method signature does not conform to the
            [`ControlLoop`][academy.agent.ControlLoop] protocol.
    """
    method._agent_method_type = 'loop'  # type: ignore[attr-defined]

    if sys.version_info >= (3, 10):  # pragma: >=3.10 cover
        found_sig = inspect.signature(method, eval_str=True)
        expected_sig = inspect.signature(ControlLoop.__call__, eval_str=True)
    else:  # pragma: <3.10 cover
        found_sig = inspect.signature(method)
        expected_sig = inspect.signature(ControlLoop.__call__)

    if found_sig != expected_sig:
        raise TypeError(
            f'Signature of loop method "{method.__name__}" is {found_sig} '
            f'but should be {expected_sig}. If the signatures look the same '
            'except that types are stringified, try importing '
            '"from __future__ import annotations" at the top of the module '
            'where the agent is defined.',
        )

    @functools.wraps(method)
    async def _wrapped(self: AgentT, shutdown: asyncio.Event) -> None:
        logger.debug('Started %r loop for %s', method.__name__, self)
        await method(self, shutdown)
        logger.debug('Exited %r loop for %s', method.__name__, self)

    return _wrapped


def event(
    name: str,
) -> Callable[
    [Callable[[AgentT], Coroutine[None, None, None]]],
    LoopMethod[AgentT],
]:
    """Decorator that annotates a method of a agent as an event loop.

    An event loop is a special type of control loop that runs when a
    [`asyncio.Event`][asyncio.Event] is set. The event is cleared
    after the loop runs.

    Example:
        ```python
        import asyncio
        from academy.agent import Agent, timer

        class Example(Agent):
            def __init__(self) -> None:
                self.alert = asyncio.Event()

            @event('alert')
            async def handle(self) -> None:
                # Runs every time alter is set
                ...
        ```

    Args:
        name: Attribute name of the [`asyncio.Event`][asyncio.Event]
            to wait on.

    Raises:
        AttributeError: Raised at runtime if no attribute named `name`
            exists on the agent.
        TypeError: Raised at runtime if the attribute named `name` is not
            a [`asyncio.Event`][asyncio.Event].
    """

    def decorator(
        method: Callable[[AgentT], Coroutine[None, None, None]],
    ) -> LoopMethod[AgentT]:
        method._agent_method_type = 'loop'  # type: ignore[attr-defined]

        @functools.wraps(method)
        async def _wrapped(self: AgentT, shutdown: asyncio.Event) -> None:
            event = getattr(self, name)
            if not isinstance(event, asyncio.Event):
                raise TypeError(
                    f'Attribute {name} of {type(self).__class__} has type '
                    f'{type(event).__class__}. Expected asyncio.Event.',
                )

            logger.debug(
                'Started %r event loop for %s (event: %r)',
                method.__name__,
                self,
                name,
            )
            while not shutdown.is_set():
                await wait_event_async(shutdown, event)
                if event.is_set():
                    try:
                        await method(self)
                    finally:
                        event.clear()
            logger.debug('Exited %r event loop for %s', method.__name__, self)

        return _wrapped

    return decorator


def timer(
    interval: float | timedelta,
) -> Callable[
    [Callable[[AgentT], Coroutine[None, None, None]]],
    LoopMethod[AgentT],
]:
    """Decorator that annotates a method of a agent as a timer loop.

    A timer loop is a special type of control loop that runs at a set
    interval. The method will always be called once before the first
    sleep.

    Example:
        ```python
        from academy.agent import Agent, timer

        class Example(Agent):
            @timer(interval=1)
            async def listen(self) -> None:
                # Runs every 1 second
                ...
        ```

    Args:
        interval: Seconds or a [`timedelta`][datetime.timedelta] to wait
            between invoking the method.
    """
    interval = (
        interval.total_seconds()
        if isinstance(interval, timedelta)
        else interval
    )

    def decorator(
        method: Callable[[AgentT], Coroutine[None, None, None]],
    ) -> LoopMethod[AgentT]:
        method._agent_method_type = 'loop'  # type: ignore[attr-defined]

        @functools.wraps(method)
        async def _wrapped(self: AgentT, shutdown: asyncio.Event) -> None:
            logger.debug(
                'Started %r timer loop for %s (interval: %fs)',
                method.__name__,
                self,
                interval,
            )
            while not shutdown.is_set():
                try:
                    await asyncio.wait_for(shutdown.wait(), timeout=interval)
                except asyncio.TimeoutError:
                    await method(self)
            logger.debug('Exited %r timer loop for %s', method.__name__, self)

        return _wrapped

    return decorator


def _is_agent_method_type(obj: Any, kind: str) -> bool:
    return (
        callable(obj)
        and hasattr(obj, '_agent_method_type')
        and obj._agent_method_type == kind
    )



from __future__ import annotations

from collections.abc import AsyncGenerator
from concurrent.futures import ThreadPoolExecutor
from typing import Any
from typing import Callable

import pytest
import pytest_asyncio
from aiohttp.web import AppRunner
from aiohttp.web import TCPSite

from academy.exchange import ExchangeFactory
from academy.exchange import HttpExchangeFactory
from academy.exchange import HybridExchangeFactory
from academy.exchange import LocalExchangeFactory
from academy.exchange import LocalExchangeTransport
from academy.exchange import RedisExchangeFactory
from academy.exchange import UserExchangeClient
from academy.exchange.cloud.server import create_app
from academy.manager import Manager
from academy.socket import open_port


@pytest_asyncio.fixture
async def http_exchange_factory(
    http_exchange_server: tuple[str, int],
) -> HttpExchangeFactory:
    host, port = http_exchange_server
    url = f'http://{host}:{port}'
    return HttpExchangeFactory(url)


@pytest.fixture
def hybrid_exchange_factory(mock_redis) -> HybridExchangeFactory:
    return HybridExchangeFactory(redis_host='localhost', redis_port=0)


@pytest.fixture
def redis_exchange_factory(mock_redis) -> RedisExchangeFactory:
    return RedisExchangeFactory(hostname='localhost', port=0)


@pytest.fixture
def local_exchange_factory() -> LocalExchangeFactory:
    return LocalExchangeFactory()


EXCHANGE_FACTORY_TYPES = (
    HttpExchangeFactory,
    HybridExchangeFactory,
    RedisExchangeFactory,
    LocalExchangeFactory,
)


@pytest_asyncio.fixture
async def get_factory(
    http_exchange_server: tuple[str, int],
    mock_redis,
) -> Callable[[type[ExchangeFactory[Any]]], ExchangeFactory[Any]]:
    # Typically we would parameterize fixtures on a list of the
    # factory fixtures defined above. However, request.getfixturevalue does
    # not work with async fixtures, of which we have a mix, so we need to set
    # them up manually. Instead, we have a fixture that returns a function
    # that can create the factories from a parameterized list of factory types.
    # See: https://github.com/pytest-dev/pytest-asyncio/issues/976
    def _get_factory_for_testing(
        factory_type: type[ExchangeFactory[Any]],
    ) -> ExchangeFactory[Any]:
        if factory_type is HttpExchangeFactory:
            host, port = http_exchange_server
            url = f'http://{host}:{port}'
            return HttpExchangeFactory(url)
        elif factory_type is HybridExchangeFactory:
            return HybridExchangeFactory(redis_host='localhost', redis_port=0)
        elif factory_type is RedisExchangeFactory:
            return RedisExchangeFactory(hostname='localhost', port=0)
        elif factory_type is LocalExchangeFactory:
            return LocalExchangeFactory()
        else:
            raise AssertionError('Unsupported factory type.')

    return _get_factory_for_testing


@pytest_asyncio.fixture
async def exchange_client() -> AsyncGenerator[
    UserExchangeClient[LocalExchangeTransport]
]:
    factory = LocalExchangeFactory()
    async with await factory.create_user_client() as client:
        yield client


@pytest.fixture
async def manager(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> AsyncGenerator[Manager[LocalExchangeTransport]]:
    async with Manager(exchange_client, ThreadPoolExecutor()) as manager:
        yield manager


@pytest_asyncio.fixture
async def http_exchange_server() -> AsyncGenerator[tuple[str, int]]:
    host, port = 'localhost', open_port()
    app = create_app()

    runner = AppRunner(app)
    await runner.setup()

    try:
        site = TCPSite(runner, host, port)
        await site.start()
        yield host, port
    finally:
        await runner.cleanup()



from __future__ import annotations

import contextlib
import enum
import sys
from collections.abc import Iterable
from types import TracebackType
from typing import Any
from typing import Protocol
from typing import runtime_checkable
from typing import TYPE_CHECKING
from typing import TypeVar

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import Self
else:  # pragma: <3.11 cover
    from typing_extensions import Self

from academy.agent import Agent
from academy.agent import AgentT
from academy.exception import MailboxTerminatedError
from academy.identifier import AgentId
from academy.identifier import EntityId
from academy.message import ErrorResponse
from academy.message import Message

if TYPE_CHECKING:
    from academy.exchange.factory import ExchangeFactory


class MailboxStatus(enum.Enum):
    """Exchange mailbox status."""

    MISSING = 'MISSING'
    """Mailbox does not exist."""
    ACTIVE = 'ACTIVE'
    """Mailbox exists and is accepting messages."""
    TERMINATED = 'TERMINATED'
    """Mailbox was terminated and no longer accepts messages."""


@runtime_checkable
class AgentRegistration(Protocol[AgentT]):
    """Agent exchange registration information.

    Attributes:
        agent_id: Unique agent identifier returned by the exchange.
    """

    agent_id: AgentId[AgentT]


AgentRegistrationT = TypeVar(
    'AgentRegistrationT',
    bound=AgentRegistration[Any],
)
"""Type variable bound [`AgentRegistration`][academy.exchange.transport.AgentRegistration]."""  # noqa: E501
AgentRegistrationT_co = TypeVar('AgentRegistrationT_co', covariant=True)


@runtime_checkable
class ExchangeTransport(Protocol[AgentRegistrationT_co]):
    """Low-level exchange communicator.

    A message exchange hosts mailboxes for each entity (i.e., agent or
    user) in a multi-agent system. This transport protocol defines mechanisms
    for entity management (e.g., registration, discovery, status, termination)
    and for sending/receiving messages from a mailbox. As such, each transport
    instance is "bound" to a specific mailbox in the exchange.

    Warning:
        A specific exchange transport should not be replicated because multiple
        client instances receiving from the same mailbox produces undefined
        agent.
    """

    @property
    def mailbox_id(self) -> EntityId:
        """ID of the mailbox this client is bound to."""
        ...

    async def close(self) -> None:
        """Close the exchange client.

        Note:
            This does not alter the state of the mailbox this client is bound
            to. I.e., the mailbox will not be terminated.
        """
        ...

    async def discover(
        self,
        agent: type[Agent],
        *,
        allow_subclasses: bool = True,
    ) -> tuple[AgentId[Any], ...]:
        """Discover peer agents with a given agent.

        Warning:
            Implementations of this method are often O(n) and scan the types
            of all agents registered to the exchange.

        Args:
            agent: Agent type of interest.
            allow_subclasses: Return agents implementing subclasses of the
                agent.

        Returns:
            Tuple of agent IDs implementing the agent.

        Raises:
            ExchangeError: Error returned by the exchange.
        """
        ...

    def factory(self) -> ExchangeFactory[Self]:
        """Get an exchange factory."""
        ...

    async def recv(self, timeout: float | None = None) -> Message[Any]:
        """Receive the next message sent to the mailbox.

        This blocks until the next message is received, there is a timeout, or
        the mailbox is terminated.

        Args:
            timeout: Optional timeout in seconds to wait for the next
                message. If `None`, the default, block forever until the
                next message or the mailbox is closed.

        Raises:
            MailboxTerminatedError: If the mailbox was closed.
            ExchangeError: Error returned by the exchange.
            TimeoutError: If a `timeout` was specified and exceeded.
        """
        ...

    async def register_agent(
        self,
        agent: type[AgentT],
        *,
        name: str | None = None,
    ) -> AgentRegistrationT_co:
        """Register a new agent and associated mailbox with the exchange.

        Args:
            agent: Agent type of the agent.
            name: Optional display name for the agent.

        Returns:
            Agent registration info.

        Raises:
            ExchangeError: Error returned by the exchange.
        """
        ...

    async def send(self, message: Message[Any]) -> None:
        """Send a message to a mailbox.

        Args:
            message: Message to send.

        Raises:
            BadEntityIdError: If a mailbox for `message.dest` does not exist.
            MailboxTerminatedError: If the mailbox was closed.
            ExchangeError: Error returned by the exchange.
        """
        ...

    async def status(self, uid: EntityId) -> MailboxStatus:
        """Check the status of a mailbox in the exchange.

        Args:
            uid: Entity identifier of the mailbox to check.

        Raises:
            ExchangeError: Error returned by the exchange.
        """
        ...

    async def terminate(self, uid: EntityId) -> None:
        """Terminate a mailbox in the exchange.

        Once an entity's mailbox is terminated:

        * All request messages in the mailbox will be replied to with a
          [`MailboxTerminatedError`][academy.exception.MailboxTerminatedError].
        * All calls to
          [`recv()`][academy.exchange.transport.ExchangeTransport.recv]
          will raise a
          [`MailboxTerminatedError`][academy.exception.MailboxTerminatedError].
        * All attempts to
          [`send()`][academy.exchange.transport.ExchangeTransport.send]
          to this mailbox by other entities will raise a
          [`MailboxTerminatedError`][academy.exception.MailboxTerminatedError].

        Note:
            This method is a no-op if the mailbox does not exist.

        Args:
            uid: Entity identifier of the mailbox to close.

        Raises:
            ExchangeError: Error returned by the exchange.
        """
        ...


ExchangeTransportT = TypeVar(
    'ExchangeTransportT',
    bound=ExchangeTransport[Any],
)
"""Type variable bound [`ExchangeTransport`][academy.exchange.transport.ExchangeTransport]."""  # noqa: E501


class ExchangeTransportMixin:
    """Magic method mixin for exchange transport implementations.

    Adds `__repr__`, `__str__`, and context manager support.
    """

    def __repr__(self: ExchangeTransportT) -> str:
        return f'{type(self).__name__}({self.mailbox_id!r})'

    def __str__(self: ExchangeTransportT) -> str:
        return f'{type(self).__name__}<{self.mailbox_id}>'

    async def __aenter__(self: ExchangeTransportT) -> ExchangeTransportT:
        return self

    async def __aexit__(
        self: ExchangeTransportT,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        await self.close()


async def _respond_pending_requests_on_terminate(
    messages: Iterable[Message[Any]],
    transport: ExchangeTransport[Any],
) -> None:
    # Helper function used to parse all pending messages in a mailbox when
    # it is terminated and reply to only request messages with a
    # MailboxTerminatedError.
    for message in messages:
        if message.is_request():
            error = MailboxTerminatedError(transport.mailbox_id)
            response = message.create_response(ErrorResponse(exception=error))
            # If the requester's mailbox was also terminated then they
            # don't need to get a response.
            with contextlib.suppress(MailboxTerminatedError):
                await transport.send(response)



from __future__ import annotations

import asyncio
from concurrent.futures import ThreadPoolExecutor

import pytest

from academy.agent import Agent
from academy.exception import BadEntityIdError
from academy.exchange import LocalExchangeFactory
from academy.exchange import LocalExchangeTransport
from academy.exchange import UserExchangeClient
from academy.manager import Manager
from testing.agents import EmptyAgent
from testing.agents import SleepAgent
from testing.constant import TEST_CONNECTION_TIMEOUT
from testing.constant import TEST_SLEEP_INTERVAL
from testing.constant import TEST_WAIT_TIMEOUT


@pytest.mark.asyncio
async def test_from_exchange_factory() -> None:
    async with await Manager.from_exchange_factory(
        factory=LocalExchangeFactory(),
        executors=ThreadPoolExecutor(),
    ) as manager:
        assert isinstance(repr(manager), str)
        assert isinstance(str(manager), str)


@pytest.mark.asyncio
async def test_launch_and_shutdown(
    manager: Manager[LocalExchangeTransport],
) -> None:
    # Two ways to launch: agent instance or deferred agent initialization
    handle1 = await manager.launch(SleepAgent(TEST_SLEEP_INTERVAL))
    handle2 = await manager.launch(SleepAgent, args=(TEST_SLEEP_INTERVAL,))

    assert len(manager.running()) == 2  # noqa: PLR2004

    await asyncio.sleep(5 * TEST_SLEEP_INTERVAL)

    await manager.shutdown(handle1.agent_id)
    await manager.shutdown(handle2)

    await manager.wait({handle1})
    await manager.wait({handle2})

    assert len(manager.running()) == 0

    # Should be a no-op since the agent is already shutdown
    await manager.shutdown(handle1)


@pytest.mark.asyncio
async def test_shutdown_on_exit(
    manager: Manager[LocalExchangeTransport],
) -> None:
    agent = SleepAgent(TEST_SLEEP_INTERVAL)
    # The manager fixture uses a context manager that when exited should
    # shutdown this agent and wait on it
    await manager.launch(agent)
    await asyncio.sleep(5 * TEST_SLEEP_INTERVAL)
    assert len(manager.running()) == 1


@pytest.mark.asyncio
async def test_wait_empty_iterable(
    manager: Manager[LocalExchangeTransport],
) -> None:
    await manager.wait({})


@pytest.mark.asyncio
async def test_wait_bad_identifier(
    manager: Manager[LocalExchangeTransport],
) -> None:
    registration = await manager.register_agent(EmptyAgent)
    with pytest.raises(BadEntityIdError):
        await manager.wait({registration.agent_id})


@pytest.mark.asyncio
async def test_wait_timeout(
    manager: Manager[LocalExchangeTransport],
) -> None:
    agent = SleepAgent(TEST_SLEEP_INTERVAL)
    handle = await manager.launch(agent)
    with pytest.raises(TimeoutError):
        await manager.wait({handle}, timeout=TEST_SLEEP_INTERVAL)


@pytest.mark.asyncio
async def test_wait_timeout_all_completed(
    manager: Manager[LocalExchangeTransport],
) -> None:
    agent = SleepAgent(TEST_SLEEP_INTERVAL)
    handle1 = await manager.launch(agent)
    handle2 = await manager.launch(agent)
    await manager.shutdown(handle1, blocking=True)
    with pytest.raises(TimeoutError):
        await manager.wait(
            {handle1, handle2},
            timeout=TEST_SLEEP_INTERVAL,
            return_when=asyncio.ALL_COMPLETED,
        )


@pytest.mark.asyncio
async def test_shutdown_bad_identifier(
    manager: Manager[LocalExchangeTransport],
) -> None:
    registration = await manager.register_agent(EmptyAgent)
    with pytest.raises(BadEntityIdError):
        await manager.shutdown(registration.agent_id)


@pytest.mark.asyncio
async def test_duplicate_launched_agents_error(
    manager: Manager[LocalExchangeTransport],
) -> None:
    registration = await manager.register_agent(EmptyAgent)
    await manager.launch(EmptyAgent(), registration=registration)
    with pytest.raises(
        RuntimeError,
        match=f'{registration.agent_id} has already been executed.',
    ):
        await manager.launch(EmptyAgent(), registration=registration)
    assert len(manager.running()) == 1


@pytest.mark.asyncio
async def test_shutdown_nonblocking(
    manager: Manager[LocalExchangeTransport],
) -> None:
    agent = SleepAgent(TEST_SLEEP_INTERVAL)
    handle = await manager.launch(agent)
    await manager.shutdown(handle, blocking=False)
    await manager.wait({handle}, timeout=TEST_WAIT_TIMEOUT)


@pytest.mark.asyncio
async def test_shutdown_blocking(
    manager: Manager[LocalExchangeTransport],
) -> None:
    agent = SleepAgent(TEST_SLEEP_INTERVAL)
    handle = await manager.launch(agent)
    await manager.shutdown(handle, blocking=True)
    assert len(manager.running()) == 0


@pytest.mark.asyncio
async def test_bad_default_executor(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    with pytest.raises(ValueError, match='No executor named "second"'):
        Manager(
            exchange_client=exchange_client,
            executors={'first': ThreadPoolExecutor()},
            default_executor='second',
        )


@pytest.mark.asyncio
async def test_add_and_set_executor_errors(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    executor = ThreadPoolExecutor()
    async with Manager(
        exchange_client=exchange_client,
        executors={'first': executor},
    ) as manager:
        with pytest.raises(
            ValueError,
            match='Executor named "first" already exists.',
        ):
            manager.add_executor('first', executor)
        with pytest.raises(
            ValueError,
            match='An executor named "second" does not exist.',
        ):
            manager.set_default_executor('second')


@pytest.mark.asyncio
async def test_multiple_executor(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    async with Manager(
        exchange_client=exchange_client,
        executors={'first': ThreadPoolExecutor()},
    ) as manager:
        await manager.launch(EmptyAgent(), executor='first')

        manager.add_executor('second', ThreadPoolExecutor())
        manager.set_default_executor('second')
        await manager.launch(EmptyAgent())
        await manager.launch(EmptyAgent(), executor='first')


@pytest.mark.asyncio
async def test_multiple_executor_no_default(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    async with Manager(
        exchange_client=exchange_client,
        executors={'first': ThreadPoolExecutor()},
    ) as manager:
        with pytest.raises(ValueError, match='no default is set.'):
            await manager.launch(EmptyAgent())


class FailOnStartupAgent(Agent):
    def __init__(self, max_errors: int | None = None) -> None:
        self.errors = 0
        self.max_errors = max_errors

    async def agent_on_startup(self) -> None:
        if self.max_errors is None or self.errors < self.max_errors:
            self.errors += 1
            raise RuntimeError('Agent startup failed')


@pytest.mark.asyncio
async def test_retry_on_error(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    # Note: this test presently relies on agent to be shared across
    # each agent execution in other threads.
    agent = FailOnStartupAgent(max_errors=2)
    executor = ThreadPoolExecutor(max_workers=1)
    async with Manager(
        exchange_client,
        executors=executor,
        max_retries=3,
    ) as manager:
        handle = await manager.launch(agent)
        await handle.ping(timeout=TEST_CONNECTION_TIMEOUT)
        assert agent.errors == 2  # noqa: PLR2004
        await handle.shutdown()


@pytest.mark.parametrize('raise_error', (True, False))
@pytest.mark.asyncio
async def test_wait_ignore_agent_errors(
    raise_error: bool,
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    agent = FailOnStartupAgent()
    manager = Manager(
        exchange_client,
        executors=ThreadPoolExecutor(max_workers=1),
    )
    handle = await manager.launch(agent)

    if raise_error:
        with pytest.raises(RuntimeError, match='Agent startup failed'):
            await manager.wait({handle}, raise_error=raise_error)
    else:
        await manager.wait({handle}, raise_error=raise_error)

    with pytest.raises(RuntimeError, match='Agent startup failed'):
        await manager.close()


@pytest.mark.asyncio
async def test_warn_executor_overload(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    agent = SleepAgent(TEST_SLEEP_INTERVAL)
    async with Manager(
        exchange_client,
        executors=ThreadPoolExecutor(max_workers=1),
    ) as manager:
        await manager.launch(agent)
        with pytest.warns(RuntimeWarning, match='Executor overload:'):
            await manager.launch(agent)
        assert len(manager.running()) == 2  # noqa: PLR2004



from __future__ import annotations

import pathlib

import pytest

from academy.agent import action
from academy.agent import Agent
from academy.state import FileState


class _StatefulAgent(Agent):
    def __init__(self, state_path: pathlib.Path) -> None:
        self.state_path = state_path

    async def agent_on_startup(self) -> None:
        self.state: FileState[str] = FileState(self.state_path)

    async def agent_on_shutdown(self) -> None:
        self.state.close()

    @action
    async def get_state(self, key: str) -> str:
        return self.state[key]

    @action
    async def modify_state(self, key: str, value: str) -> None:
        self.state[key] = value


@pytest.mark.asyncio
async def test_file_state(tmp_path: pathlib.Path) -> None:
    state_path = tmp_path / 'state.dbm'

    agent = _StatefulAgent(state_path)
    await agent.agent_on_startup()
    key, value = 'foo', 'bar'
    await agent.modify_state(key, value)
    assert await agent.get_state(key) == value
    await agent.agent_on_shutdown()

    agent = _StatefulAgent(state_path)
    await agent.agent_on_startup()
    assert await agent.get_state(key) == value
    await agent.agent_on_shutdown()



# ruff: noqa: D102
from __future__ import annotations

import asyncio
import base64
import contextlib
import dataclasses
import logging
import sys
import uuid
from collections.abc import Iterable
from typing import Any
from typing import Generic

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import Self
else:  # pragma: <3.11 cover
    from typing_extensions import Self

if sys.version_info >= (3, 13):  # pragma: >=3.13 cover
    from asyncio import Queue
    from asyncio import QueueShutDown

    AsyncQueue = Queue
else:  # pragma: <3.13 cover
    # Use of queues here is isolated to a single thread/event loop so
    # we only need culsans queues for the backport of shutdown() agent
    from culsans import AsyncQueue
    from culsans import AsyncQueueShutDown as QueueShutDown
    from culsans import Queue

import redis.asyncio

from academy.agent import Agent
from academy.agent import AgentT
from academy.exception import BadEntityIdError
from academy.exception import MailboxTerminatedError
from academy.exchange.factory import ExchangeFactory
from academy.exchange.redis import _MailboxState
from academy.exchange.redis import _RedisConnectionInfo
from academy.exchange.transport import _respond_pending_requests_on_terminate
from academy.exchange.transport import ExchangeTransportMixin
from academy.exchange.transport import MailboxStatus
from academy.identifier import AgentId
from academy.identifier import EntityId
from academy.identifier import UserId
from academy.message import Message
from academy.serialize import NoPickleMixin
from academy.socket import address_by_hostname
from academy.socket import address_by_interface
from academy.socket import open_port
from academy.socket import SimpleSocketServer
from academy.socket import SocketClosedError
from academy.socket import SocketPool

logger = logging.getLogger(__name__)

_CLOSE_SENTINEL = b'<CLOSED>'
_THREAD_START_TIMEOUT = 5
_THREAD_JOIN_TIMEOUT = 5
_SERVER_ACK = b'<ACK>'
_SOCKET_POLL_TIMEOUT_MS = 50


@dataclasses.dataclass
class HybridAgentRegistration(Generic[AgentT]):
    """Agent registration for hybrid exchanges."""

    agent_id: AgentId[AgentT]
    """Unique identifier for the agent created by the exchange."""


class HybridExchangeTransport(ExchangeTransportMixin, NoPickleMixin):
    """Hybrid exchange transport bound to a specific mailbox."""

    def __init__(  # noqa: PLR0913
        self,
        mailbox_id: EntityId,
        redis_client: redis.asyncio.Redis,
        *,
        redis_info: _RedisConnectionInfo,
        namespace: str,
        host: str,
        port: int,
        interface: str | None = None,
    ) -> None:
        self._mailbox_id = mailbox_id
        self._redis_client = redis_client
        self._redis_info = redis_info
        self._namespace = namespace
        self._host = host
        self._port = port
        self._interface = interface

        self._address_cache: dict[EntityId, str] = {}
        if sys.version_info >= (3, 13):  # pragma: >=3.13 cover
            self._messages: AsyncQueue[Message[Any]] = Queue()
        else:  # pragma: <3.13 cover
            self._messages: AsyncQueue[Message[Any]] = Queue().async_q
        self._socket_pool = SocketPool()
        self._started = asyncio.Event()
        self._shutdown = asyncio.Event()

        self._server = SimpleSocketServer(
            handler=self._direct_message_handler,
            host=host,
            port=port,
        )
        self._server_task = asyncio.create_task(
            self._run_direct_server(),
            name=f'hybrid-transport-direct-server-{self.mailbox_id}',
        )
        self._redis_task = asyncio.create_task(
            self._run_redis_listener(),
            name=f'hybrid-transport-redis-watcher-{self.mailbox_id}',
        )

    def _address_key(self, uid: EntityId) -> str:
        return f'{self._namespace}:address:{uuid_to_base32(uid.uid)}'

    def _agent_key(self, aid: AgentId[Any]) -> str:
        return f'{self._namespace}:agent:{uuid_to_base32(aid.uid)}'

    def _status_key(self, uid: EntityId) -> str:
        return f'{self._namespace}:status:{uuid_to_base32(uid.uid)}'

    def _queue_key(self, uid: EntityId) -> str:
        return f'{self._namespace}:queue:{uuid_to_base32(uid.uid)}'

    @classmethod
    async def new(  # noqa: PLR0913
        cls,
        *,
        namespace: str,
        redis_info: _RedisConnectionInfo,
        interface: str | None = None,
        mailbox_id: EntityId | None = None,
        name: str | None = None,
        port: int | None = None,
    ) -> Self:
        """Instantiate a new transport.

        Args:
            namespace: Redis key namespace.
            redis_info: Redis connection information.
            interface: Network interface use for peer-to-peer communication.
                If `None`, the hostname of the local host is used.
            mailbox_id: Bind the transport to the specific mailbox. If `None`,
                a new user entity will be registered and the transport will be
                bound to that mailbox.
            name: Display name of the registered entity if `mailbox_id` is
                `None`.
            port: Port to listen for peer connection on.

        Returns:
            An instantiated transport bound to a specific mailbox.

        Raises:
            redis.exceptions.ConnectionError: If the Redis server is not
                reachable.
        """
        host = (
            address_by_interface(interface)
            if interface is not None
            else address_by_hostname()
        )
        port = port if port is not None else open_port()

        client = redis.asyncio.Redis(
            host=redis_info.hostname,
            port=redis_info.port,
            decode_responses=False,
            **redis_info.kwargs,
        )
        # Ensure the redis server is reachable else fail early
        await client.ping()

        if mailbox_id is None:
            mailbox_id = UserId.new(name=name)
            await client.set(
                f'{namespace}:status:{uuid_to_base32(mailbox_id.uid)}',
                _MailboxState.ACTIVE.value,
            )
            logger.info('Registered %s in exchange', mailbox_id)

        await client.set(
            f'{namespace}:address:{uuid_to_base32(mailbox_id.uid)}',
            f'{host}:{port}',
        )

        transport = cls(
            mailbox_id,
            client,
            redis_info=redis_info,
            namespace=namespace,
            interface=interface,
            host=host,
            port=port,
        )
        # Wait for the direct message server to start
        await asyncio.wait_for(transport._started.wait(), timeout=5)
        return transport

    @property
    def mailbox_id(self) -> EntityId:
        return self._mailbox_id

    async def close(self) -> None:
        self._shutdown.set()
        self._messages.shutdown(immediate=True)
        await self._socket_pool.close()

        await asyncio.wait_for(self._server_task, timeout=5)

        self._redis_task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await self._redis_task

        await self._redis_client.aclose()

    async def discover(
        self,
        agent: type[Agent],
        *,
        allow_subclasses: bool = True,
    ) -> tuple[AgentId[Any], ...]:
        found: list[AgentId[Any]] = []
        fqp = f'{agent.__module__}.{agent.__name__}'
        async for key in self._redis_client.scan_iter(  # pragma: no branch
            f'{self._namespace}:agent:*',
        ):
            mro_str = await self._redis_client.get(key)
            assert isinstance(mro_str, str)
            mro = mro_str.split(',')
            if fqp == mro[0] or (allow_subclasses and fqp in mro):
                aid: AgentId[Any] = AgentId(
                    uid=base32_to_uuid(key.split(':')[-1]),
                )
                found.append(aid)
        active: list[AgentId[Any]] = []
        for aid in found:
            status = await self._redis_client.get(self._status_key(aid))
            if status == _MailboxState.ACTIVE.value:  # pragma: no branch
                active.append(aid)
        return tuple(active)

    def factory(self) -> HybridExchangeFactory:
        return HybridExchangeFactory(
            redis_host=self._redis_info.hostname,
            redis_port=self._redis_info.port,
            redis_kwargs=self._redis_info.kwargs,
            interface=self._interface,
            namespace=self._namespace,
        )

    async def recv(self, timeout: float | None = None) -> Message[Any]:
        try:
            return await asyncio.wait_for(
                self._messages.get(),
                timeout=timeout,
            )
        except asyncio.TimeoutError:
            raise TimeoutError(
                f'Timeout waiting for next message for {self.mailbox_id} '
                f'after {timeout} seconds.',
            ) from None
        except QueueShutDown:
            raise MailboxTerminatedError(self.mailbox_id) from None

    async def register_agent(
        self,
        agent: type[AgentT],
        *,
        name: str | None = None,
    ) -> HybridAgentRegistration[AgentT]:
        aid: AgentId[AgentT] = AgentId.new(name=name)
        await self._redis_client.set(
            self._status_key(aid),
            _MailboxState.ACTIVE.value,
        )
        await self._redis_client.set(
            self._agent_key(aid),
            ','.join(agent._agent_mro()),
        )
        return HybridAgentRegistration(agent_id=aid)

    async def _send_direct(self, address: str, message: Message[Any]) -> None:
        await self._socket_pool.send(address, message.model_serialize())
        logger.debug(
            'Sent %s to %s via p2p at %s',
            type(message).__name__,
            message.dest,
            address,
        )

    async def send(self, message: Message[Any]) -> None:
        address = self._address_cache.get(message.dest, None)
        if address is not None:
            try:
                # This is as optimistic as possible. If the address of the
                # peer is cached, we assume the mailbox is still active and
                # the peer is still listening.
                await self._send_direct(address, message)
            except (SocketClosedError, OSError):
                # Our optimism let us down so clear the cache and try the
                # standard flow.
                self._address_cache.pop(message.dest)
            else:
                return

        status = await self._redis_client.get(self._status_key(message.dest))
        if status is None:
            raise BadEntityIdError(message.dest)
        elif status == _MailboxState.INACTIVE.value:
            raise MailboxTerminatedError(message.dest)

        maybe_address = await self._redis_client.get(
            self._address_key(message.dest),
        )
        try:
            # This branching is a little odd. We want to fall back to
            # Redis for message sending on two conditions: direct send fails
            # or no address was found. We raise a TypeError if no address
            # was found as a shortcut to get to the fall back.
            if isinstance(maybe_address, (bytes, str)):
                decoded_address = (
                    maybe_address.decode('utf-8')
                    if isinstance(maybe_address, bytes)
                    else maybe_address
                )
                await self._send_direct(decoded_address, message)
                self._address_cache[message.dest] = decoded_address
            else:
                raise TypeError('Did not active peer address in Redis.')
        except (TypeError, SocketClosedError, OSError):
            await self._redis_client.rpush(  # type: ignore[misc]
                self._queue_key(message.dest),
                message.model_serialize(),
            )
            logger.debug(
                'Sent %s to %s via redis',
                type(message).__name__,
                message.dest,
            )

    async def status(self, uid: EntityId) -> MailboxStatus:
        status = await self._redis_client.get(self._status_key(uid))
        if status is None:
            return MailboxStatus.MISSING
        elif status == _MailboxState.INACTIVE.value:
            return MailboxStatus.TERMINATED
        else:
            return MailboxStatus.ACTIVE

    async def terminate(self, uid: EntityId) -> None:
        # Warning: terminating a hybrid exchange mailbox is not guaranteed
        # to respond to all pending requests with a MailboxTerminatedError
        # when messages are sent directly.
        await self._redis_client.set(
            self._status_key(uid),
            _MailboxState.INACTIVE.value,
        )

        pending = await self._redis_client.lrange(self._queue_key(uid), 0, -1)  # type: ignore[misc]
        await self._redis_client.delete(self._queue_key(uid))
        # Sending a close sentinel to the queue is a quick way to force
        # the entity waiting on messages to the mailbox to stop blocking.
        # This assumes that only one entity is reading from the mailbox.
        await self._redis_client.rpush(self._queue_key(uid), _CLOSE_SENTINEL)  # type: ignore[misc]

        messages: list[Message[Any]] = [
            Message.model_deserialize(raw) for raw in pending
        ]
        await _respond_pending_requests_on_terminate(messages, self)

        if isinstance(uid, AgentId):
            await self._redis_client.delete(self._agent_key(uid))

    async def _get_message_from_redis(self) -> None:
        # Block indefinitely with timeout=0
        raw = await self._redis_client.blpop(  # type: ignore[misc]
            [self._queue_key(self.mailbox_id)],
            timeout=0,
        )

        # Only passed one key to blpop to result is [key, item]
        assert isinstance(raw, (tuple, list))
        assert len(raw) == 2  # noqa: PLR2004
        if raw[1] == _CLOSE_SENTINEL:  # pragma: no cover
            self._shutdown.set()
            self._messages.shutdown(immediate=True)
            raise MailboxTerminatedError(self.mailbox_id)
        message: Message[Any] = Message.model_deserialize(raw[1])
        logger.debug(
            'Received %s to %s via redis',
            type(message).__name__,
            self.mailbox_id,
        )
        await self._messages.put(message)

    async def _run_redis_listener(self) -> None:
        logger.debug('Started redis listener task for %s', self.mailbox_id)
        try:
            while True:
                status = await self._redis_client.get(
                    self._status_key(self.mailbox_id),
                )
                if status is None:  # pragma: no cover
                    raise AssertionError(
                        f'Status for mailbox {self.mailbox_id} did not exist '
                        'in Redis server. This means that something '
                        'incorrectly deleted the key.',
                    )
                elif (
                    status == _MailboxState.INACTIVE.value
                ):  # pragma: no cover
                    break
                await self._get_message_from_redis()
        except MailboxTerminatedError:  # pragma: no cover
            pass
        except Exception:
            logger.exception(
                'Error in redis listener task for %s',
                self.mailbox_id,
            )
        finally:
            self._shutdown.set()
            self._messages.shutdown(immediate=True)
            logger.debug(
                'Stopped redis listener task for %s',
                self.mailbox_id,
            )

    async def _direct_message_handler(self, payload: bytes) -> bytes | None:
        message: Message[Any] = Message.model_deserialize(payload)
        logger.debug(
            'Received %s to %s via p2p',
            type(message).__name__,
            self.mailbox_id,
        )
        await self._messages.put(message)
        return None

    async def _run_direct_server(self) -> None:
        logger.debug(
            'Started direct message server task for %s',
            self.mailbox_id,
        )
        async with self._server.serve():
            self._started.set()
            await self._shutdown.wait()

            await self._redis_client.delete(
                self._address_key(self.mailbox_id),
            )


class HybridExchangeFactory(ExchangeFactory[HybridExchangeTransport]):
    """Hybrid exchange client factory.

    The hybrid exchange uses peer-to-peer communication via TCP and a
    central Redis server for mailbox state and queueing messages for
    offline entities.

    Args:
        redis_host: Redis server hostname.
        redis_port: Redis server port.
        redis_kwargs: Extra keyword arguments to pass to
            [`redis.Redis()`][redis.Redis].
        interface: Network interface use for peer-to-peer communication. If
            `None`, the hostname of the local host is used.
        namespace: Redis key namespace. If `None` a random key prefix is
            generated.
        ports: An iterable of ports to give each client a unique port from a
            user defined set. A StopIteration exception will be raised in
            `create_*_client()` methods if the number of clients in the process
            is greater than the length of the iterable.
    """

    def __init__(  # noqa: PLR0913
        self,
        redis_host: str,
        redis_port: int,
        *,
        redis_kwargs: dict[str, Any] | None = None,
        interface: str | None = None,
        namespace: str | None = 'default',
        ports: Iterable[int] | None = None,
    ) -> None:
        self._namespace = (
            namespace
            if namespace is not None
            else uuid_to_base32(uuid.uuid4())
        )
        self._interface = interface
        self._redis_info = _RedisConnectionInfo(
            redis_host,
            redis_port,
            redis_kwargs if redis_kwargs is not None else {},
        )
        self._ports = None if ports is None else iter(ports)

    async def _create_transport(
        self,
        mailbox_id: EntityId | None = None,
        *,
        name: str | None = None,
        registration: HybridAgentRegistration[Any] | None = None,  # type: ignore[override]
    ) -> HybridExchangeTransport:
        return await HybridExchangeTransport.new(
            interface=self._interface,
            mailbox_id=mailbox_id,
            name=name,
            namespace=self._namespace,
            port=None if self._ports is None else next(self._ports),
            redis_info=self._redis_info,
        )


def base32_to_uuid(uid: str) -> uuid.UUID:
    """Parse a base32 string as a UUID."""
    padding = '=' * ((8 - len(uid) % 8) % 8)
    padded = uid + padding
    uid_bytes = base64.b32decode(padded)
    return uuid.UUID(bytes=uid_bytes)


def uuid_to_base32(uid: uuid.UUID) -> str:
    """Encode a UUID as a trimmed base32 string."""
    uid_bytes = uid.bytes
    base32_bytes = base64.b32encode(uid_bytes).rstrip(b'=')
    return base32_bytes.decode('utf-8')



from __future__ import annotations

from academy.exchange.cloud.client import HttpExchangeFactory
from academy.exchange.cloud.client import HttpExchangeTransport
from academy.exchange.cloud.client import spawn_http_exchange

__all__ = [
    'HttpExchangeFactory',
    'HttpExchangeTransport',
    'spawn_http_exchange',
]



from __future__ import annotations

import asyncio
import logging
import multiprocessing
from concurrent.futures import ProcessPoolExecutor

from academy.agent import action
from academy.agent import Agent
from academy.exchange.cloud import spawn_http_exchange
from academy.handle import Handle
from academy.logging import init_logging
from academy.manager import Manager

EXCHANGE_PORT = 5346
logger = logging.getLogger(__name__)


class Coordinator(Agent):
    def __init__(
        self,
        lowerer: Handle[Lowerer],
        reverser: Handle[Reverser],
    ) -> None:
        super().__init__()
        self.lowerer = lowerer
        self.reverser = reverser

    @action
    async def process(self, text: str) -> str:
        text = await self.lowerer.lower(text)
        text = await self.reverser.reverse(text)
        return text


class Lowerer(Agent):
    @action
    async def lower(self, text: str) -> str:
        return text.lower()


class Reverser(Agent):
    @action
    async def reverse(self, text: str) -> str:
        return text[::-1]


async def main() -> int:
    init_logging(logging.INFO)

    with spawn_http_exchange('localhost', EXCHANGE_PORT) as factory:
        mp_context = multiprocessing.get_context('spawn')
        executor = ProcessPoolExecutor(
            max_workers=3,
            initializer=init_logging,
            mp_context=mp_context,
        )

        async with await Manager.from_exchange_factory(
            factory=factory,
            # Agents are run by the manager in the processes of this
            # process pool executor.
            executors=executor,
        ) as manager:
            # Launch each of the three agents types. The returned type is
            # a handle to that agent used to invoke actions.
            lowerer = await manager.launch(Lowerer)
            reverser = await manager.launch(Reverser)
            coordinator = await manager.launch(
                Coordinator,
                args=(lowerer, reverser),
            )

            text = 'DEADBEEF'
            expected = 'feebdaed'

            logger.info(
                'Invoking process("%s") on %s',
                text,
                coordinator.agent_id,
            )
            result = await coordinator.process(text)
            assert result == expected
            logger.info('Received result: "%s"', result)

        # Upon exit, the Manager context will instruct each agent to shutdown,
        # closing their respective handles, and shutting down the executors.

    return 0


if __name__ == '__main__':
    raise SystemExit(asyncio.run(main()))



from __future__ import annotations

import abc
import asyncio
import contextlib
import logging
import sys
import uuid
from collections.abc import Coroutine
from types import TracebackType
from typing import Any
from typing import Callable
from typing import Generic
from typing import TYPE_CHECKING
from weakref import WeakValueDictionary

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import TypeAlias
else:  # pragma: <3.11 cover
    from typing_extensions import TypeAlias

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import Self
else:  # pragma: <3.11 cover
    from typing_extensions import Self

from academy.agent import Agent
from academy.agent import AgentT
from academy.exception import MailboxTerminatedError
from academy.exchange.transport import AgentRegistration
from academy.exchange.transport import ExchangeTransportT
from academy.exchange.transport import MailboxStatus
from academy.handle import exchange_context
from academy.handle import Handle
from academy.identifier import AgentId
from academy.identifier import EntityId
from academy.identifier import UserId
from academy.message import ErrorResponse
from academy.message import Message
from academy.message import RequestT_co

if TYPE_CHECKING:
    from academy.exchange.factory import ExchangeFactory

logger = logging.getLogger(__name__)

RequestHandler: TypeAlias = Callable[
    [Message[RequestT_co]],
    Coroutine[None, None, None],
]


class ExchangeClient(abc.ABC, Generic[ExchangeTransportT]):
    """Base exchange client.

    Warning:
        Exchange clients should only be created via
        [`ExchangeFactory.create_agent_client()`][academy.exchange.ExchangeFactory.create_agent_client]
        or
        [`ExchangeFactory.create_user_client()`][academy.exchange.ExchangeFactory.create_user_client]!

    Args:
        transport: Exchange transport bound to a mailbox.
    """

    def __init__(
        self,
        transport: ExchangeTransportT,
    ) -> None:
        self._transport = transport
        self._handles: WeakValueDictionary[uuid.UUID, Handle[Any]] = (
            WeakValueDictionary()
        )
        self._close_lock = asyncio.Lock()
        self._closed = False

    async def __aenter__(self) -> Self:
        self.exchange_context_token = exchange_context.set(self)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        exchange_context.reset(self.exchange_context_token)
        await self.close()

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self.client_id!r})'

    def __str__(self) -> str:
        return f'{type(self).__name__}<{self.client_id}>'

    @property
    @abc.abstractmethod
    def client_id(self) -> EntityId:
        """Client ID as registered with the exchange."""
        ...

    @abc.abstractmethod
    async def close(self) -> None:
        """Close the transport."""
        ...

    async def discover(
        self,
        agent: type[Agent],
        *,
        allow_subclasses: bool = True,
    ) -> tuple[AgentId[Any], ...]:
        """Discover peer agents with a given agent.

        Args:
            agent: Agent type of interest.
            allow_subclasses: Return agents implementing subclasses of the
                agent.

        Returns:
            Tuple of agent IDs implementing the agent.
        """
        return await self._transport.discover(
            agent,
            allow_subclasses=allow_subclasses,
        )

    def factory(self) -> ExchangeFactory[ExchangeTransportT]:
        """Get an exchange factory."""
        return self._transport.factory()

    def register_handle(self, handle: Handle[AgentT]) -> None:
        """Register an existing handle to receive messages.

        Args:
            handle: Handle to register.
        """
        self._handles[handle.handle_id] = handle

    async def register_agent(
        self,
        agent: type[AgentT],
        *,
        name: str | None = None,
    ) -> AgentRegistration[AgentT]:
        """Register a new agent and associated mailbox with the exchange.

        Args:
            agent: Agent type of the agent.
            name: Optional display name for the agent.

        Returns:
            Agent registration info.
        """
        registration = await self._transport.register_agent(
            agent,
            name=name,
        )
        logger.info('Registered %s in exchange', registration.agent_id)
        return registration

    async def send(self, message: Message[Any]) -> None:
        """Send a message to a mailbox.

        Args:
            message: Message to send.

        Raises:
            BadEntityIdError: If a mailbox for `message.dest` does not exist.
            MailboxTerminatedError: If the mailbox was closed.
        """
        await self._transport.send(message)
        logger.debug('Sent %s to %s', type(message).__name__, message.dest)

    async def status(self, uid: EntityId) -> MailboxStatus:
        """Check the status of a mailbox in the exchange.

        Args:
            uid: Entity identifier of the mailbox to check.
        """
        return await self._transport.status(uid)

    async def terminate(self, uid: EntityId) -> None:
        """Terminate a mailbox in the exchange.

        Terminating a mailbox means that the corresponding entity will no
        longer be able to receive messages.

        Note:
            This method is a no-op if the mailbox does not exist.

        Args:
            uid: Entity identifier of the mailbox to close.
        """
        await self._transport.terminate(uid)

    async def _listen_for_messages(self) -> None:
        while True:
            try:
                message = await self._transport.recv()
            except (asyncio.CancelledError, MailboxTerminatedError):
                break
            logger.debug(
                'Received %s from %s for %s',
                type(message).__name__,
                message.src,
                self.client_id,
            )
            await self._handle_message(message)

    @abc.abstractmethod
    async def _handle_message(self, message: Message[Any]) -> None: ...


class AgentExchangeClient(
    ExchangeClient[ExchangeTransportT],
    Generic[AgentT, ExchangeTransportT],
):
    """Agent exchange client.

    Warning:
        Agent exchange clients should only be created via
        [`ExchangeFactory.create_agent_client()`][academy.exchange.ExchangeFactory.create_agent_client]!

    Args:
        agent_id: Agent ID.
        transport: Exchange transport bound to `agent_id`.
        request_handler: Request handler of the agent that will be called
            for each message received to this agent's mailbox.
            start_listener: Start a message listener thread.
    """

    def __init__(
        self,
        agent_id: AgentId[AgentT],
        transport: ExchangeTransportT,
        request_handler: RequestHandler[RequestT_co],
    ) -> None:
        super().__init__(transport)
        self._agent_id = agent_id
        self._request_handler = request_handler

    @property
    def client_id(self) -> AgentId[AgentT]:
        """Agent ID of the client."""
        return self._agent_id

    async def close(self) -> None:
        """Close the user client.

        This closes the underlying exchange transport and all handles created
        by this client. The agent's mailbox will not be terminated so the agent
        can be started again later.
        """
        async with self._close_lock:
            if self._closed:
                return

            await self._transport.close()
            self._closed = True
            logger.info('Closed exchange client for %s', self.client_id)

    async def _handle_message(self, message: Message[Any]) -> None:
        if message.is_request():
            await self._request_handler(message)
        elif message.is_response():
            if message.label is None or message.label not in self._handles:
                logger.warning(
                    'Exchange client for %s received an unexpected response '
                    'message from %s but no corresponding handle exists.',
                    self.client_id,
                    message.src,
                )
                return
            handle = self._handles[message.label]
            await handle._process_response(message)
        else:
            raise AssertionError('Unreachable.')


class UserExchangeClient(ExchangeClient[ExchangeTransportT]):
    """User exchange client.

    Warning:
        User exchange clients should only be created via
        [`ExchangeFactory.create_user_client()`][academy.exchange.ExchangeFactory.create_user_client]!

    Args:
        user_id: User ID.
        transport: Exchange transport bound to `user_id`.
        start_listener: Start a message listener thread.
    """

    def __init__(
        self,
        user_id: UserId,
        transport: ExchangeTransportT,
        *,
        start_listener: bool = True,
    ) -> None:
        super().__init__(transport)
        self._user_id = user_id
        self._listener_task: asyncio.Task[None] | None = None
        if start_listener:
            self._listener_task = asyncio.create_task(
                self._listen_for_messages(),
                name=f'user-exchange-listener-{self.client_id}',
            )

    @property
    def client_id(self) -> UserId:
        """User ID of the client."""
        return self._user_id

    async def close(self) -> None:
        """Close the user client.

        This terminates the user's mailbox, closes the underlying exchange
        transport.
        """
        async with self._close_lock:
            if self._closed:
                return

            await self._transport.terminate(self.client_id)
            logger.info(f'Terminated mailbox for {self.client_id}')
            await self._stop_listener_task()
            await self._transport.close()
            self._closed = True
            logger.info('Closed exchange client for %s', self.client_id)

    async def _handle_message(self, message: Message[Any]) -> None:
        if message.is_request():
            error = TypeError(f'{self.client_id} cannot fulfill requests.')
            response = message.create_response(ErrorResponse(exception=error))
            await self._transport.send(response)
            logger.warning(
                'Exchange client for %s received unexpected request message '
                'from %s',
                self.client_id,
                message.src,
            )
        elif message.is_response():
            if (
                message.label is None or message.label not in self._handles
            ):  # pragma: no cover
                logger.warning(
                    'Exchange client for %s received an unexpected response '
                    'message from %s but no corresponding handle exists.',
                    self.client_id,
                    message.src,
                )
                return
            handle = self._handles[message.label]
            await handle._process_response(message)
        else:
            raise AssertionError('Unreachable.')

    async def _stop_listener_task(self) -> None:
        if self._listener_task is not None:
            self._listener_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._listener_task
            self._listener_task = None



from __future__ import annotations

import sys
import uuid
from typing import Any
from typing import Generic
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import Self
else:  # pragma: <3.11 cover
    from typing_extensions import Self

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field

# Normally this would be bound=Agent, but Pydantic's mypy plugin crashes
# here. See https://github.com/pydantic/pydantic/issues/11454
AgentT = TypeVar('AgentT')


class AgentId(BaseModel, Generic[AgentT]):
    """Unique identifier for an agent entity in a multi-agent system."""

    uid: uuid.UUID = Field()
    name: Optional[str] = Field(None)  # noqa: UP045
    role: Literal['agent'] = Field('agent', repr=False)

    model_config = ConfigDict(
        extra='forbid',
        frozen=True,
        validate_default=True,
    )

    def __eq__(self, other: object, /) -> bool:
        return isinstance(other, AgentId) and self.uid == other.uid

    def __hash__(self) -> int:
        return hash(self.role) + hash(self.uid)

    def __str__(self) -> str:
        name = self.name if self.name is not None else str(self.uid)[:8]
        return f'AgentId<{name}>'

    @classmethod
    def new(cls, name: str | None = None) -> Self:
        """Create a new identifier.

        Args:
            name: Optional human-readable name for the entity.
        """
        return cls(uid=uuid.uuid4(), name=name)


class UserId(BaseModel):
    """Unique identifier for a user entity in a multi-agent system."""

    uid: uuid.UUID = Field()
    name: Optional[str] = Field(None)  # noqa: UP045
    role: Literal['user'] = Field('user', repr=False)

    model_config = ConfigDict(
        extra='forbid',
        frozen=True,
        validate_default=True,
    )

    def __eq__(self, other: object, /) -> bool:
        return isinstance(other, UserId) and self.uid == other.uid

    def __hash__(self) -> int:
        return hash(self.role) + hash(self.uid)

    def __str__(self) -> str:
        name = self.name if self.name is not None else str(self.uid)[:8]
        return f'UserId<{name}>'

    @classmethod
    def new(cls, name: str | None = None) -> Self:
        """Create a new identifier.

        Args:
            name: Optional human-readable name for the entity.
        """
        return cls(uid=uuid.uuid4(), name=name)


if TYPE_CHECKING:
    EntityId = Union[AgentId[Any], UserId]
    """EntityId union type for type annotations."""
else:
    # Pydantic produces validation errors with Agent[Any] in versions
    # prior to 2.10.0. We could require academy to use newer versions of
    # pydantic but this could be a headache for third-party apps with
    # stricter requirements.
    # Issue: https://github.com/pydantic/pydantic/issues/9414
    # Fix: https://github.com/pydantic/pydantic/pull/10666
    EntityId = Union[AgentId, UserId]



from __future__ import annotations

import pickle
from collections.abc import Generator
from typing import Any
from typing import Callable

import pytest
from proxystore.connectors.local import LocalConnector
from proxystore.proxy import Proxy
from proxystore.store import Store
from proxystore.store.executor import ProxyAlways
from proxystore.store.executor import ProxyNever

from academy.exchange import MailboxStatus
from academy.exchange.local import LocalExchangeFactory
from academy.exchange.proxystore import ProxyStoreExchangeFactory
from academy.message import ActionRequest
from academy.message import ActionResponse
from academy.message import Message
from academy.message import PingRequest
from testing.agents import EmptyAgent


@pytest.fixture
def store() -> Generator[Store[LocalConnector], None, None]:
    with Store(
        'proxystore-exchange-store-fixture',
        LocalConnector(),
        cache_size=0,
        register=True,
    ) as store:
        yield store


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ('should_proxy', 'resolve_async'),
    (
        (ProxyNever(), False),
        (ProxyAlways(), True),
        (ProxyAlways(), False),
        (lambda x: isinstance(x, str), True),
    ),
)
async def test_wrap_basic_transport_functionality(
    should_proxy: Callable[[Any], bool],
    resolve_async: bool,
    store: Store[LocalConnector],
    local_exchange_factory: LocalExchangeFactory,
) -> None:
    wrapped_factory = ProxyStoreExchangeFactory(
        base=local_exchange_factory,
        store=store,
        should_proxy=should_proxy,
        resolve_async=resolve_async,
    )

    async with await wrapped_factory._create_transport() as wrapped_transport1:
        new_factory = wrapped_transport1.factory()
        assert isinstance(new_factory, ProxyStoreExchangeFactory)

        src = wrapped_transport1.mailbox_id
        dest = (await wrapped_transport1.register_agent(EmptyAgent)).agent_id
        assert await wrapped_transport1.status(dest) == MailboxStatus.ACTIVE

        wrapped_transport2 = await wrapped_factory._create_transport(
            mailbox_id=dest,
        )
        assert wrapped_transport2.mailbox_id == dest

        ping = Message.create(src=src, dest=dest, body=PingRequest())
        await wrapped_transport1.send(ping)
        assert await wrapped_transport2.recv() == ping

        sent_request = ActionRequest(
            action='test',
            pargs=('value', 123),
            kargs={'foo': 'value', 'bar': 123},
        )
        sent_request_message = Message.create(
            src=src,
            dest=dest,
            body=sent_request,
        )
        await wrapped_transport1.send(sent_request_message)

        recv_request_message = await wrapped_transport2.recv()
        recv_request = recv_request_message.get_body()
        assert isinstance(recv_request, ActionRequest)
        assert sent_request_message.tag == recv_request_message.tag

        for old, new in zip(sent_request.get_args(), recv_request.get_args()):
            assert (type(new) is Proxy) == should_proxy(old)
            # will resolve the proxy if it exists
            assert old == new

        for name in recv_request.get_kwargs():
            old = sent_request.get_kwargs()[name]
            new = recv_request.get_kwargs()[name]
            assert (type(new) is Proxy) == should_proxy(old)
            assert old == new

        sent_response = ActionResponse(action='test', result='result')
        sent_response_message = sent_request_message.create_response(
            sent_response,
        )
        await wrapped_transport2.send(sent_response_message)

        recv_response_message = await wrapped_transport1.recv()
        recv_response = recv_response_message.get_body()
        assert isinstance(recv_response, ActionResponse)
        assert sent_response_message.tag == recv_response_message.tag
        assert (type(recv_response.get_result()) is Proxy) == should_proxy(
            sent_response.get_result(),
        )
        assert sent_response.get_result() == recv_response.get_result()

        assert await wrapped_transport1.discover(EmptyAgent) == (dest,)

        await wrapped_transport1.terminate(wrapped_transport1.mailbox_id)
        await wrapped_transport2.close()


@pytest.mark.asyncio
async def test_serialize_factory(
    http_exchange_factory,
    store: Store[LocalConnector],
) -> None:
    wrapped_factory = ProxyStoreExchangeFactory(
        base=http_exchange_factory,
        store=store,
        should_proxy=ProxyAlways(),
    )
    dumped = pickle.dumps(wrapped_factory)
    reconstructed = pickle.loads(dumped)
    assert isinstance(reconstructed, ProxyStoreExchangeFactory)



# ruff: noqa: D102
from __future__ import annotations

import functools
from collections.abc import Iterable
from collections.abc import Mapping
from typing import Any
from typing import Callable
from typing import Generic

from proxystore.proxy import Proxy
from proxystore.store import get_or_create_store
from proxystore.store import register_store
from proxystore.store import Store
from proxystore.store.utils import resolve_async

from academy.agent import Agent
from academy.agent import AgentT
from academy.exchange.factory import ExchangeFactory
from academy.exchange.transport import AgentRegistration
from academy.exchange.transport import AgentRegistrationT
from academy.exchange.transport import ExchangeTransportMixin
from academy.exchange.transport import ExchangeTransportT
from academy.exchange.transport import MailboxStatus
from academy.identifier import AgentId
from academy.identifier import EntityId
from academy.message import ActionRequest
from academy.message import ActionResponse
from academy.message import Message
from academy.serialize import NoPickleMixin


def _proxy_item(
    item: Any,
    store: Store[Any],
    should_proxy: Callable[[Any], bool],
) -> Any:
    if type(item) is not Proxy and should_proxy(item):
        return store.proxy(item)
    return item


def _proxy_iterable(
    items: Iterable[Any],
    store: Store[Any],
    should_proxy: Callable[[Any], bool],
) -> tuple[Any, ...]:
    _apply = functools.partial(
        _proxy_item,
        store=store,
        should_proxy=should_proxy,
    )
    return tuple(map(_apply, items))


def _proxy_mapping(
    mapping: Mapping[Any, Any],
    store: Store[Any],
    should_proxy: Callable[[Any], bool],
) -> dict[Any, Any]:
    _apply = functools.partial(
        _proxy_item,
        store=store,
        should_proxy=should_proxy,
    )
    return {key: _apply(item) for key, item in mapping.items()}


class ProxyStoreExchangeTransport(
    ExchangeTransportMixin,
    NoPickleMixin,
    Generic[ExchangeTransportT],
):
    """ProxyStore exchange client bound to a specific mailbox."""

    def __init__(
        self,
        transport: ExchangeTransportT,
        store: Store[Any],
        should_proxy: Callable[[Any], bool],
        *,
        resolve_async: bool = False,
    ) -> None:
        self.transport = transport
        self.store = store
        self.should_proxy = should_proxy
        self.resolve_async = resolve_async
        register_store(store, exist_ok=True)

    @property
    def mailbox_id(self) -> EntityId:
        return self.transport.mailbox_id

    async def close(self) -> None:
        await self.transport.close()

    async def discover(
        self,
        agent: type[Agent],
        *,
        allow_subclasses: bool = True,
    ) -> tuple[AgentId[Any], ...]:
        return await self.transport.discover(
            agent,
            allow_subclasses=allow_subclasses,
        )

    def factory(self) -> ProxyStoreExchangeFactory[ExchangeTransportT]:
        return ProxyStoreExchangeFactory(
            self.transport.factory(),
            self.store,
            should_proxy=self.should_proxy,
            resolve_async=self.resolve_async,
        )

    async def recv(self, timeout: float | None = None) -> Message[Any]:
        message = await self.transport.recv(timeout)
        body = message.get_body()
        if self.resolve_async and isinstance(body, ActionRequest):
            args = body.get_args()
            kwargs = body.get_kwargs()
            for arg in (*args, *kwargs.values()):
                if type(arg) is Proxy:
                    resolve_async(arg)
        elif (
            self.resolve_async
            and isinstance(body, ActionResponse)
            and type(body.get_result()) is Proxy
        ):
            resolve_async(body.get_result())
        return message

    async def register_agent(
        self,
        agent: type[AgentT],
        *,
        name: str | None = None,
    ) -> AgentRegistration[AgentT]:
        return await self.transport.register_agent(agent, name=name)

    async def send(self, message: Message[Any]) -> None:
        body = message.get_body()
        if isinstance(body, ActionRequest):
            body.pargs = _proxy_iterable(
                body.get_args(),
                self.store,
                self.should_proxy,
            )
            body.kargs = _proxy_mapping(
                body.get_kwargs(),
                self.store,
                self.should_proxy,
            )
        elif isinstance(body, ActionResponse):
            body.result = _proxy_item(
                body.result,
                self.store,
                self.should_proxy,
            )

        await self.transport.send(message)

    async def status(self, uid: EntityId) -> MailboxStatus:
        return await self.transport.status(uid)

    async def terminate(self, uid: EntityId) -> None:
        await self.transport.terminate(uid)


class ProxyStoreExchangeFactory(
    ExchangeFactory[ProxyStoreExchangeTransport[ExchangeTransportT]],
):
    """ProxStore exchange client factory.

    A ProxyStore exchange is used to wrap an underlying exchange so
    large objects may be passed by reference.

    Args:
        base: Base exchange factory.
        store: Store to use for proxying data.
        should_proxy: A callable that returns `True` if an object should be
            proxied. This is applied to every positional and keyword argument
            and result value.
        resolve_async: Resolve proxies asynchronously when received.
    """

    def __init__(
        self,
        base: ExchangeFactory[ExchangeTransportT],
        store: Store[Any] | None,
        should_proxy: Callable[[Any], bool],
        *,
        resolve_async: bool = False,
    ) -> None:
        self.base = base
        self.store = store
        self.should_proxy = should_proxy
        self.resolve_async = resolve_async

    async def _create_transport(
        self,
        mailbox_id: EntityId | None = None,
        *,
        name: str | None = None,
        registration: AgentRegistrationT | None = None,
    ) -> ProxyStoreExchangeTransport[ExchangeTransportT]:
        # If store was none because of pickling,
        # the __setstate__ must be called before bind.
        assert self.store is not None
        transport = await self.base._create_transport(
            mailbox_id,
            name=name,
            registration=registration,
        )
        return ProxyStoreExchangeTransport(
            transport,
            self.store,
            self.should_proxy,
            resolve_async=self.resolve_async,
        )

    def __getstate__(self) -> dict[str, Any]:
        assert self.store is not None

        return {
            'base': self.base,
            'store_config': self.store.config(),
            'resolve_async': self.resolve_async,
            'should_proxy': self.should_proxy,
        }

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.store = get_or_create_store(
            state.pop('store_config'),
            register=True,
        )
        self.__dict__.update(state)



from __future__ import annotations

import pickle

import pytest

from academy.serialize import NoPickleMixin


class CanItPickle(NoPickleMixin):
    pass


def test_no_pickle_mixin() -> None:
    with pytest.raises(pickle.PicklingError):
        pickle.dumps(CanItPickle())
